# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_main_wing_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE 
from RCAIDE.Framework.Core    import Units

# python imports 
import  numpy as  np
import  copy 
 
# ----------------------------------------------------------------------------------------------------------------------
# Main Wing Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_wing_weight(vehicle, wing, WPOD, complexity, settings, num_main_wings):
    """ Calculate the wing weight based on the flops method. The wing weight consists of:
        - Total Wing Shear Material and Control Surface Weight
        - Total Wing Miscellaneous Items Weight
        - Total Wing Bending Material Weight

        Assumptions:
            Wing is elliptically loaded
            Gloved wing area is 0
            Load between multiple main wings is distributed equally
            Wing sweep is fixed

        Source:
            The Flight Optimization System Weight Estimation Method

       Inputs:
            vehicle - data dictionary with vehicle properties                   [dimensionless]
                -.reference_area: wing surface area                             [m^2]
                -.mass_properties.max_takeoff: MTOW                             [kilograms]
                -.flight_envelope.ultimate_load: ultimate load factor (default: 3.75)
                -.systems.accessories: type of aircraft (short-range, commuter
                                                        medium-range, long-range,
                                                        sst, cargo)
                -.fuselages.fuselage.width: width of the fuselage               [m]
             -wing: data dictionary with wing properties
                    -.taper: taper ratio
                    -.sweeps.quarter_chord: quarter chord sweep angle           [deg]
                    -.thickness_to_chord: thickness to chord
                    -.spans.projected: wing span                                [m]
                    -.chords.root: root chord                                   [m]
                    -.tip.root: tip chord                                       [m]
                    -.twists.root: twist of wing at root                        [deg]
                    -.twists.tip: twist of wing at tip                          [deg]
                    -.flap_ratio: flap surface area over wing surface area
                 -.networks: data dictionary containing all propulsion properties
                    -.number_of_engines: number of engines
                    -.sealevel_static_thrust: thrust at sea level               [N]
            WPOD - weight of engine pod including the nacelle                   [kilograms]
            complexity - "simple" or "complex" depending on the wing weight method chosen

       Outputs:
            WWING - wing weight                                          [kilograms]

        Properties Used:
            N/A
    """
    SW          = wing.areas.reference / (Units.ft ** 2)  # Reference wing area, ft^2
    GLOV        = 0 
    SX          = SW - GLOV  # Wing trapezoidal area
    SPAN        = wing.spans.projected / Units.ft  # Wing span, ft
    SEMISPAN    = SPAN / 2
    AR          = SPAN ** 2 / SX  # Aspect ratio
    TR          = wing.taper  # Taper
    
    aeroelastic_tailoring_factor = settings.FLOPS.aeroelastic_tailoring_factor
    strut_braced_wing_factor     = settings.FLOPS.strut_braced_wing_factor
    composite_utilization_factor = settings.FLOPS.composite_utilization_factor
    
    if AR <= 5:
        CAYA = 0
    else:
        CAYA = AR - 5
    # Aeroelastic tailoring factor [0 no aeroelastic tailoring, 1 maximum aeroelastic tailoring]
    FAERT           = aeroelastic_tailoring_factor  
    # Wing strut bracing factor [0 for no struts, 1 for struts]
    FSTRT           = strut_braced_wing_factor 
    
    NENG = 0
    NEW  = 0
    for network in  vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                NENG += 1  
                if propulsor.wing_mounted: 
                    NEW += 1
                        
    DG              = vehicle.mass_properties.max_takeoff / Units.lbs  # Design gross weight in lb

    if complexity == 'Simple':
        EMS  = 1 - 0.25 * FSTRT  # Wing strut bracing factor
        TLAM = np.tan(wing.sweeps.quarter_chord) \
               - 2 * (1 - TR) / (AR * (1 + TR))  # Tangent of the 3/4 chord sweep angle
        SLAM = TLAM / np.sqrt(1 + TLAM ** 2)  # sine of 3/4 chord wing sweep angle
        C6   = 0.5 * FAERT - 0.16 * FSTRT
        C4   = 1 - 0.5 * FAERT
        CAYL = (1.0 - SLAM ** 2) * \
               (1.0 + C6 * SLAM ** 2 + 0.03 * CAYA * C4 * SLAM)  # Wing sweep factor due to aeroelastic tailoring
        TCA  = wing.thickness_to_chord
        BT   = 0.215 * (0.37 + 0.7 * TR) * (SPAN ** 2 / SW) ** EMS / (CAYL * TCA)  # Bending factor
        CAYE = 1 - 0.03 * NEW

    else:
        NSD             = 500
        N2              = int(NEW / 2) 
        L_fus           = 0
        for fuselage in vehicle.fuselages:
            if L_fus < fuselage.lengths.total:
                ref_fuselage = fuselage 
        ETA, C, T, SWP  = generate_wing_stations(ref_fuselage.width, copy.deepcopy(wing))
        NS, Y           = generate_int_stations(NSD, ETA)
        EETA            = get_spanwise_engine(vehicle.networks,SEMISPAN)
        P0              = calculate_load(ETA[-1])
        ASW             = 0
        EM              = 0
        EL              = 0
        C0              = C[-1]
        S               = 0
        EEL             = 0
        EEM             = 0
        EA0             = 0
        EW              = 0
        
        
        # Replaced FOR LOOP
        # Reverse Order
        Y  = np.flip(Y)
        
        # DY distance
        DY = np.diff(Y)
        
        # Trim the vectors away from the tip and center
        Y  = Y[1:-2]
        DY = -DY[0:-2]
        
        # Get normalized pressure loading across the wing
        P1     = calculate_load(Y)
        P0     = np.zeros_like(P1)
        P0[1:] = P1[0:-1]
        
        # Get local chord length
        C1     = np.interp(Y, ETA, C)
        C0     = np.zeros_like(C1)
        C0[0]  = C[-1]
        C0[1:] = C1[0:-1]
        
        # Calculate local pressure load and moments (DELP and DELM)
        T1   = np.interp(Y, ETA, T)
        SWP1 = find_sweep(Y,ETA,SWP)
        DELP = DY / 6 * (C0 * (2 * P0 + P1) + C1 * (2 * P1 + P0))
        DELM = DY ** 2 * (C0 * (3.0 * P0 + P1) + C1 * (P1 + P0)) / 12.
        
        # Sum loads
        EL     = np.zeros_like(DELP) 
        EL[1:] = np.cumsum(DELP[0:-1])
        
        # Sum moments
        EM     = np.cumsum((DELM + DY * EL) * 1 / np.cos(SWP1))
        
        # Calculate required bending material area
        BMA1     = EM * 1 / np.cos(SWP1) * 1 / (C1 * T1)
        
        BMA0     = np.zeros_like(BMA1)
        BMA0[1:] = BMA1[0:-1]
        
        # Compute segment values
        ASW  = np.cumsum((DY + 2 * Y) * DY * SWP1)
        PM   = np.cumsum((BMA0 + BMA1) * DY / 2.)
        S    = np.cumsum((C0 + C1) * DY / 2.)


        # Adjust for engine loads
        if N2>0: # If there are engines
            EEL   = np.zeros_like(Y)
            DELM2 = np.zeros_like(Y)
            
            # Do a for loop over engine stations
            for ii in range(len(EETA)):
                # Find the station closest to the engine but inboard
                distances              = EETA[ii]-Y
                distances[distances<0] = np.inf
                distance               = np.min(distances)
                loc                    = np.argmin(distances)
                DELM2[loc]             = DELM2[loc] + distance
                EEL[loc+1:]            = EEL[loc+1:] + 1

            DELM2 = DELM2 + EEL*DY

            EEM = np.cumsum(DELM2/np.cos(SWP1))
            EA1 = EEM * 1 / np.cos(SWP1) * 1 / (C1 * T1)
            
            EA0 = np.zeros_like(Y)
            EA0[1:] = EA1[0:-1]
            
            EW  = np.sum((EA0 + EA1) * DY / 2)
            
        # Finalize properties
        EL = EL[-1] + DELP[-1]    
        EM = EM[-1] / EL
        PM = 4. * PM[-1] / EL
        EW = 8. * EW
        SA = np.sin(ASW[-1])
        AR = 2 / S[-1]       
                
        if AR <= 5:
            CAYA = 0
        else:
            CAYA = AR - 5
        DEN = AR ** (.25 * FSTRT) * (1.0 + (.50 * FAERT - .160 * FSTRT) * SA ** 2 /
                                     + .03 * CAYA * (1.0 - .50 * FAERT) * SA)
        BT = PM / DEN
        BTE = EW
        CAYE = 1
        if NEW > 0:
            CAYE = 1 - BTE / BT * WPOD / DG

    A       = wing_weight_constants_FLOPS(vehicle.systems.accessories)  # Wing weight constants
    # Composite utilization factor [0 no composite, 1 full composite]
    FCOMP   = composite_utilization_factor  
    ULF     = vehicle.flight_envelope.ultimate_load
    if len(vehicle.fuselages) == 1:
        CAYF    = 1  # Multiple fuselage factor [1 one fuselage, 0.5 multiple fuselages]
    elif len(vehicle.fuselage) > 1:
        CAYF    = 0.5
    else:
        raise NotImplementedError
    VFACT   = 1  # Variable sweep factor, TODO: add equation to allow variable sweep penalty
    PCTL    = 1/num_main_wings  # Fraction of load carried by this wing
    W1NIR   = A[0] * BT * (1 + np.sqrt(A[1] / SPAN)) * ULF * SPAN * (1 - 0.4 * FCOMP) * (
                1 - 0.1 * FAERT) * CAYF * VFACT * PCTL / 10.0 ** 6  # Wing bending material weight lb
    SFLAP   = wing.flap_ratio * SX

    W2 = A[2] * (1 - 0.17 * FCOMP) * SFLAP ** (A[3]) * DG ** (A[4])  # shear material weight
    W3 = A[5] * (1 - 0.3 * FCOMP) * SW ** (A[6])  # miscellaneous items weight
    W1 = (DG * CAYE * W1NIR + W2 + W3) / (1 + W1NIR) - W2 - W3  # bending material weight
    WWING = W1 + W2 + W3  # Total wing weight

    return WWING * Units.lbs


def generate_wing_stations(fuselage_width, wing):
    """ Divides half the wing in sections, using the defined sections
        and adding a section at the intersection of wing and fuselage

        Assumptions:

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            fuselage_width: fuselage width                                      [m]
            wing: data dictionary with wing properties
                    -.taper: taper ration wing
                    -.sweeps.quarter_chord: quarter chord sweep angle           [deg]
                    -.thickness_to_chord: thickness to chord
                    -.spans.projected: wing span                                [m]
                    -.chords.root: root chord                                   [m]
                    -.tip.root: tip chord                                       [m]
                    -.twists.root: twist of wing at root                        [deg]
                    -.twists.tip: twist of wing at tip                          [deg]
                    -.Segments: trapezoidal segments of the wing

       Outputs:
           ETA: spanwise location of the sections normalized by half span
           C: chord lengths at every spanwise location in ETA normalized by half span
           T: thickness to chord ratio at every span wise location in ETA
           SWP: quarter chord sweep angle at every span wise location in ETA

        Properties Used:
            N/A
    """
    SPAN        = wing.spans.projected / Units.ft  # Wing span, ft
    SEMISPAN    = SPAN / 2
    root_chord  = wing.chords.root / Units.ft
    num_seg     = len(wing.segments.keys())

    if num_seg == 0:
        segment                         = RCAIDE.Library.Components.Wings.Segments.Segment()
        segment.tag                     = 'root'
        segment.percent_span_location   = 0.
        segment.twist                   = wing.twists.root
        segment.root_chord_percent      = 1
        segment.dihedral_outboard       = 0.
        segment.sweeps.quarter_chord    = wing.sweeps.quarter_chord
        segment.thickness_to_chord      = wing.thickness_to_chord
        wing.segments.append(segment)

        segment                         = RCAIDE.Library.Components.Wings.Segments.Segment()
        segment.tag                     = 'tip'
        segment.percent_span_location   = 1.
        segment.twist                   = wing.twists.tip
        segment.root_chord_percent      = wing.chords.tip / wing.chords.root
        segment.dihedral_outboard       = 0.
        segment.sweeps.quarter_chord    = wing.sweeps.quarter_chord
        segment.thickness_to_chord      = wing.thickness_to_chord
        wing.segments.append(segment)
        num_seg = len(wing.segments.keys())
        
    ETA    = np.zeros(num_seg + 1)
    C      = np.zeros(num_seg + 1)
    T      = np.zeros(num_seg + 1)
    SWP    = np.zeros(num_seg + 1)

    segment_keys  = list(wing.segments.keys())     
    ETA[0] = wing.segments[segment_keys[0]].percent_span_location
    C[0]   = root_chord * wing.segments[segment_keys[0]].root_chord_percent * 1 / SEMISPAN
    SWP[0] = 0
    
    if hasattr(wing.segments[segment_keys[0]], 'thickness_to_chord'):
        T[0] = wing.segments[segment_keys[0]].thickness_to_chord
    else:
        T[0] = wing.thickness_to_chord
    ETA[1] = fuselage_width / 2 * 1 / Units.ft * 1 / SEMISPAN
    C[1] = determine_fuselage_chord(fuselage_width, wing) * 1 / SEMISPAN

    if hasattr(wing.segments[segment_keys[0]], 'thickness_to_chord'):
        T[1] = wing.segments[segment_keys[0]].thickness_to_chord
    else:
        T[1] = wing.thickness_to_chord
    for i in range(1, num_seg):
        ETA[i + 1] = wing.segments[segment_keys[i]].percent_span_location
        C[i + 1] = root_chord * wing.segments[segment_keys[i]].root_chord_percent * 1 / SEMISPAN
        if hasattr(wing.segments[segment_keys[i]], 'thickness_to_chord'):
            T[i + 1] = wing.segments[segment_keys[i]].thickness_to_chord
        else:
            T[i + 1] = wing.thickness_to_chord
        SWP[i] = np.arctan(np.tan(wing.segments[segment_keys[i-1]].sweeps.quarter_chord) - (C[i - 1] - C[i]))
    SWP[-1] = np.arctan(np.tan(wing.segments[segment_keys[-2]].sweeps.quarter_chord) - (C[-2] - C[-1]))
    return ETA, C, T, SWP


def generate_int_stations(NSD, ETA):
    """ Divides half of the wing in integration stations

        Assumptions:

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            NSD: number of integration stations requested
            ETA: list of spanwise locations of all sections of the wing

       Outputs:
           NS: actual number of integration stations
           Y: spanwise locations of the integrations stations normalized by half span

        Properties Used:
            N/A
    """
    Y           = [ETA[1]]
    desired_int = (ETA[-1] - ETA[1]) / NSD
    NS          = 0
    for i in range(2, len(ETA)):
        NP = int((ETA[i] - ETA[i - 1]) / desired_int + 0.5)
        if NP < 1:
            NP = 1
        AINT = (ETA[i] - ETA[i - 1]) / NP
        for j in range(NP):
            NS = NS + 1
            Y.append(Y[-1] + AINT)
    return NS, Y


def calculate_load(ETA):
    """ Returns load factor assuming elliptical load distribution

        Assumptions:

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            ETA: list of spanwise locations of all sections of the wing

       Outputs:
           PS: load factor at every location in ETA assuming elliptical load distribution

        Properties Used:
            N/A
    """
    PS = np.sqrt(1. - ETA ** 2)
    return PS


def find_sweep(y, lst_y, swp):
    """ Finds sweep angle for a certain y-location along the wing

        Assumptions:

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            y: spanwise location
            lst_y: list of spanwise stations where sweep is known (eg sections)
            swp: list of quarter chord sweep angles at the locations listed in lst_y

       Outputs:
           swps: sweep angle at y

        Properties Used:
            N/A
    """
    
    # All initial sweeps are the root chord sweep
    swps = np.ones_like(y)*swp[0]
    
    for i in range(len(lst_y)-1):
        e       = lst_y[i]
        swps[y>=e] = swp[i]
        

    return swps


def get_spanwise_engine(networks, SEMISPAN):
    """ Returns EETA for the engine locations along the wing

        Assumptions:

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            networks: data dictionary with all the engine properties
                -.wing_mounted: list of boolean if engine is mounted to wing
                -.number_of_engines: number of engines
                -.origin: origin of the engine
            SEMISPAN: half span                                 [m]
       Outputs:
           EETA: span wise locations of the engines mounted to the wing normalized by the half span

        Properties Used:
            N/A
    """
    EETA =  []
    for network in  networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Propulsors.Turbojet):
                if propulsor.wing_mounted and propulsor.origin[0][1] > 0:  
                    EETA.append((propulsor.origin[0][1] / Units.ft) * 1 / SEMISPAN) 
    EETA =  np.array(EETA)
    return EETA


def wing_weight_constants_FLOPS(ac_type):
    """Defines wing weight constants as defined by FLOPS
        Inputs: ac_type - determines type of instruments, electronics, and operating items based on types:
                "short-range", "medium-range", "long-range", "business", "cargo", "commuter", "sst"
        Outputs: list of coefficients used in weight estimations

    """
    if ac_type == "short_range" or ac_type == "business" or \
            ac_type == "commuter":
        A = [30.0, 0., 0.25, 0.5, 0.5, 0.16, 1.2]
    else:
        A = [8.8, 6.25, 0.68, 0.34, 0.6, 0.035, 1.5]
    return A


def determine_fuselage_chord(fuselage_width, wing):
    """ Determine chord at wing and fuselage intersection

        Assumptions:
            Fuselage side of body is between first and second wing segments.

        Source:
            The Flight Optimization System Weight Estimation Method

        Inputs:
            fuselage_width: width of fuselage                                   [m]
            wing: data dictionary with wing properties
                    -.taper: taper ratio
                    -.sweeps.quarter_chord: quarter chord sweep angle           [deg]
                    -.thickness_to_chord: thickness to chord
                    -.spans.projected: wing span                                [m]
                    -.chords.root: root chord                                   [m]
                -.fuselages.fuselage.width: fuselage width                      [m]
       Outputs:
           chord: chord length of wing where wing intersects the fuselage wall [ft]

        Properties Used:
            N/A
    """

    segment_keys    = list(wing.segments.keys())      
    root_chord      = wing.chords.root / Units.ft
    SPAN            = wing.spans.projected / Units.ft  # Wing span, ft
    SEMISPAN        = SPAN / 2
    c1              = root_chord * wing.segments[segment_keys[0]].root_chord_percent
    c2              = root_chord * wing.segments[segment_keys[-1]].root_chord_percent
    y1              = wing.segments[segment_keys[0]].percent_span_location
    y2              = wing.segments[segment_keys[1]].percent_span_location
    b               = (y2 - y1) * SEMISPAN
    taper           = c2 / c1
    y               = fuselage_width / 2 * 1 / Units.ft
    chord           = c1 * (1 - (1 - taper) * 2 * y / b)
    return chord




# ----------------------------------------------------------------------
#  Compute Chord Length from Span Location
# ---------------------------------------------------------------------- 

## @ingroup Methods-Geometry-Three_Dimensional
def compute_chord_length_from_span_location(wing,span_location):
    """Computes the chord length given a location along the half-span.

    Assumptions:
    Linear variation of chord with span.

    Source:
    None

    Inputs:
    wing.chords.
      root                [m]
      tip                 [m]
    wing.spans.projected  [m]
    span_location         [m]

    Outputs:
    chord_length          [m]

    Properties Used:
    N/A
    """
    #unpack
    ct = wing.chords.tip
    cr = wing.chords.root
    b  = wing.spans.projected
    
    b_2 = b/2.
    
    chord_length = ct + ((cr-ct)/b_2)*(b_2-span_location)

    
    return chord_length


# RCAIDE/Library/Methods/Thermal_Management/Reservoirs/Reservoir_Tank/__init__.py

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 
from .append_reservoir_conditions   import append_reservoir_conditions
from .append_reservoir_conditions   import append_reservoir_segment_conditions
from .compute_mixing_temperature    import compute_mixing_temperature



## @ingroup Methods-Propulsion
# rayleigh.py
# 
# Created:  Aug 2017, P. Goncalves
# Modified: Jan 2018, W. Maier

import numpy as np

from scipy.optimize import fsolve

# ----------------------------------------------------------------------
#  rayleigh
# ----------------------------------------------------------------------

## @ingroup Methods-Propulsion
def rayleigh(gamma, M0, TtR):
    """
    Function that takes in a input (output) Mach number and a stagnation 
    temperature ratio and yields an output (input) Mach number, according
    to the Rayleigh flow equation. The function also outputs the stagnation
    pressure ratio
    
    Inputs:
    M       [dimensionless]
    gamma   [dimensionless]
    Ttr     [dimensionless]
    
    Outputs:
    M1      [dimensionless]
    Ptr     [dimensionless]
    
    """

    func = lambda M1: (((1.+gamma*M0*M0)**2.*M1*M1*(1.+(gamma-1.)/2.*M1*M1))/((1.+gamma*M1*M1)**2.*M0*M0*(1.+(gamma-1.)/2.*M0*M0))-TtR)

    #Initializing the array
    M1_guess = np.ones_like(M0)
    
    # Separating supersonic and subsonic solutions
    i_low = M0 <= 1.0
    i_high = M0 > 1.0

    #--Subsonic solution Guess
    M1_guess[i_low]= .01
    
    #--Supersonic solution Guess
    M1_guess[i_high]= 1.1

    # Find Mach number
    M1 = fsolve(func,M1_guess, factor=0.1)
    
    #Calculate stagnation pressure ratio
    Ptr = ((1.+gamma*M0*M0)/(1.+gamma*M1*M1)*((1.+(gamma-1.)/2.*M1*M1)/(1.+(gamma-1.)/2.*M0*M0))**(gamma/(gamma-1.)))

    return M1, Ptr


# RCAIDE/Methods/Performance/aircraft_aerodynamic_analysis.py
# 
# 
# Created:  Dec 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import  Data 
 
# Pacakge imports 
import numpy as np  

#------------------------------------------------------------------------------
# aircraft_aerodynamic_analysis
#------------------------------------------------------------------------------  
def aircraft_aerodynamic_analysis(vehicle,
                                  angle_of_attack_range,
                                  Mach_number_range,
                                  control_surface_deflection_range = np.array([[0]]),
                                  altitude = 0,
                                  delta_ISA=0,
                                  use_surrogate = True,
                                  model_fuselage = True):  

    #------------------------------------------------------------------------
    # setup flight conditions
    #------------------------------------------------------------------------   
    atmosphere     = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data      = atmosphere.compute_values(altitude,delta_ISA)
    P              = atmo_data.pressure 
    T              = atmo_data.temperature 
    rho            = atmo_data.density 
    a              = atmo_data.speed_of_sound 
    mu             = atmo_data.dynamic_viscosity
       
    # -----------------------------------------------------------------
    # Evaluate Without Surrogate
    # ----------------------------------------------------------------- 
    ctrl_pts = len(angle_of_attack_range[:, 0] )
    state                                         = RCAIDE.Framework.Mission.Common.State()
    state.conditions                              = RCAIDE.Framework.Mission.Common.Results() 
    state.conditions.freestream.density           = rho * np.ones_like(angle_of_attack_range)
    state.conditions.freestream.dynamic_viscosity = mu  * np.ones_like(angle_of_attack_range)
    state.conditions.freestream.temperature       = T   * np.ones_like(angle_of_attack_range)
    state.conditions.freestream.pressure          = P   * np.ones_like(angle_of_attack_range)
    state.conditions.aerodynamics.angles.alpha    = angle_of_attack_range  
    state.conditions.aerodynamics.angles.beta     = angle_of_attack_range *0  
    state.conditions.freestream.u                 = angle_of_attack_range *0       
    state.conditions.freestream.v                 = angle_of_attack_range *0       
    state.conditions.freestream.w                 = angle_of_attack_range *0       
    state.conditions.static_stability.roll_rate   = angle_of_attack_range *0       
    state.conditions.static_stability.pitch_rate  = angle_of_attack_range *0 
    state.conditions.static_stability.yaw_rate    = angle_of_attack_range *0  
    state.conditions.expand_rows(ctrl_pts)
 
    CL_vals    = np.zeros((len(angle_of_attack_range),len(Mach_number_range)))  
    CD_vals    = np.zeros((len(angle_of_attack_range),len(Mach_number_range)))
    
 
    state.analyses                                  =  Data()
    aerodynamics                                    = RCAIDE.Framework.Analyses.Aerodynamics.Vortex_Lattice_Method() 
    aerodynamics.settings.use_surrogate             = use_surrogate 
    aerodynamics.vehicle                            = vehicle
    aerodynamics.settings.model_fuselage            = model_fuselage   
    aerodynamics.initialize()
    state.analyses.aerodynamics = aerodynamics 
    
    for i in range (len(Mach_number_range)):  
        state.conditions.freestream.mach_number                 = Mach_number_range[i, 0] * np.ones_like(angle_of_attack_range)
        state.conditions.freestream.velocity                    = Mach_number_range[i, 0] * a   * np.ones_like(angle_of_attack_range)   
        state.conditions.freestream.reynolds_number             = state.conditions.freestream.density * state.conditions.freestream.velocity / state.conditions.freestream.dynamic_viscosity 
        state.conditions.frames.inertial.velocity_vector[:,0]   = Mach_number_range[i, 0] * a[0, 0]   *  angle_of_attack_range[:, 0] 
        
     
        # ---------------------------------------------------------------------------------------
        # Evaluate With Surrogate
        # ---------------------------------------------------------------------------------------  
        _                 = state.analyses.aerodynamics.evaluate(state)        
        CL_vals[:,i]      = state.conditions.aerodynamics.coefficients.lift.total[:, 0]
        CD_vals[:,i]      = state.conditions.aerodynamics.coefficients.drag.total[:, 0] 

  
    results = Data(
        Mach              = Mach_number_range, 
        alpha             = angle_of_attack_range, 
        lift_coefficient  = CL_vals, 
        drag_coefficient  = CD_vals, 
    )  
         
    # FUTURE WORK          
    #control_surfaces_aerodynamics = Data()
    #num_defl =  len(control_surface_deflection_range)
    #for wing in vehicle.wings:
        #for control_surface in  wing.control_surfaces: 
            #CL_cs = np.zeros((num_defl,1))
            #CD_cs = np.zeros((num_defl,1))
            #for cs_i in  range(num_defl):
                #control_surface.deflection = control_surface_deflection_range[i]
                
                #_        = state.analyses.aerodynamics.evaluate(state)     
                #CL_cs[:,cs_i]    = state.conditions.aerodynamics.coefficients.lift.total
                #CD_cs[:,cs_i]    = state.conditions.aerodynamics.coefficients.drag.total 
                
            #control_surfaces_aerodynamics[control_surface.tag].lift_coefficient =  CL_cs
            #control_surfaces_aerodynamics[control_surface.tag].drag_coefficient =  CD_cs
            
            #control_surface.deflection =  0
    #results.control_surfaces_aerodynamics    = control_surfaces_aerodynamics
    #results.control_surface_deflection_range = control_surface_deflection_range 
    return results  



# RCAIDE/Library/Methods/Thermal_Management/Heat_Exchangers/Cross_Flow_Heat_Exchanger/__init__.py

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from .design_cross_flow_heat_exchanger            import design_cross_flow_heat_exchanger
from .compute_heat_exhanger_factors               import compute_heat_exhanger_factors
from .cross_flow_heat_exchanger_sizing_setup      import cross_flow_heat_exchanger_sizing_setup
from .cross_flow_heat_exchanger_geometry_setup    import cross_flow_heat_exchanger_geometry_setup
from .cross_flow_hex_rating_model                 import cross_flow_hex_rating_model
from .append_cross_flow_heat_exchanger_conditions import append_cross_flow_heat_exchanger_conditions
from .append_cross_flow_heat_exchanger_conditions import append_cross_flow_hex_segment_conditions


# RCAIDE/Library/Compoments/Nacelles/Segment.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from RCAIDE.Framework.Core import Container
from RCAIDE.Library.Components import Component  

# ---------------------------------------------------------------------------------------------------------------------- 
#  Segment
# ----------------------------------------------------------------------------------------------------------------------   
class Segment(Component):
    """
    A component representing a single cross-sectional segment of a nacelle. Used to build
    up complex nacelle geometries through a series of connected sections.

    Attributes
    ----------
    tag : str
        Unique identifier for the segment, defaults to 'segment'
        
    orientation_euler_angles : list
        Rotation angles [roll, pitch, yaw], defaults to [0.0, 0.0, 0.0]
        
    percent_x_location : float
        Longitudinal position as percentage of nacelle length, defaults to 0
        
    percent_y_location : float
        Lateral position as percentage of nacelle width, defaults to 0
        
    percent_z_location : float
        Vertical position as percentage of nacelle height, defaults to 0
        
    height : float
        Vertical dimension of the segment cross-section, defaults to 0
        
    width : float
        Lateral dimension of the segment cross-section, defaults to 0
        
    curvature : float
        Shape parameter controlling cross-section corner rounding using super-ellipse 
        formulation, defaults to 2

    Notes
    -----
    Segments are used to build up the complete nacelle geometry through a series of 
    cross-sections. Each segment's position is defined as a percentage of the overall 
    nacelle dimensions, allowing for flexible scaling and positioning.
    
    **Major Assumptions**
    
    * Segments are connected in sequence to form a continuous surface
    * Cross-sections are defined by height, width, and curvature parameters
    * Super-ellipse formulation used for cross-section shape
    
    **Definitions**

    'Super-ellipse'
        Mathematical curve used to create smooth, rounded rectangular shapes with
        controllable corner curvature

    See Also
    --------
    RCAIDE.Library.Components.Nacelles.Stack_Nacelle
        Nacelle type that uses multiple segments
    """

    def __defaults__(self): 
        """
        Sets default values for the nacelle segment attributes.
        """
        self.tag                      = 'segment' 
        self.orientation_euler_angles = [0.,0.,0.]  
        self.percent_x_location       = 0  
        self.percent_y_location       = 0
        self.percent_z_location       = 0 
        self.height                   = 0 
        self.width                    = 0 
        self.curvature                = 2 # super ellipse 
         
class Segment_Container(Container):
    """
    Container class for managing nacelle segments. Provides organization and 
    access methods for segment components.

    Notes
    -----
    This container is designed to hold and manage Segment objects that define 
    the nacelle shape. It inherits from the base Container class and provides 
    specialized functionality for nacelle segments.

    See Also
    --------
    RCAIDE.Framework.Core.Container
        Base container class providing core functionality
    RCAIDE.Library.Components.Nacelles.Segment
        The segment components stored in this container
    """     

    def get_children(self):
        """
        Returns a list of allowable child component types for the segment container.

        Returns
        -------
        list
            Empty list as segments do not contain child components
        """       
        return []


# RCAIDE/Library/Missions/Segments/Single_Point/Set_Speed_Set_Throttle.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  

import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------   
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    A fixed speed and throttle

    Source:
    N/A

    Inputs:
    segment.altitude                               [meters]
    segment.air_speed                              [meters/second]
    segment.throttle                               [unitless]
    segment.linear_acceleration_z                         [meters/second^2]
    segment.state.unknowns.acceleration            [meters/second^2]

    Outputs:
    conditions.frames.inertial.acceleration_vector [meters/second^2]
    conditions.frames.inertial.velocity_vector     [meters/second]
    conditions.frames.inertial.position_vector     [meters]
    conditions.freestream.altitude                 [meters]
    conditions.frames.inertial.time                [seconds]

    Properties Used:
    N/A
    """      
    
    # unpack
    alt                     = segment.altitude
    air_speed               = segment.air_speed
    beta                    = segment.sideslip_angle  
    linear_acceleration_z   = segment.linear_acceleration_z
    roll_rate               = segment.roll_rate
    pitch_rate              = segment.pitch_rate
    yaw_rate                = segment.yaw_rate
    acceleration            = segment.state.unknowns.acceleration[0][0]
    
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    v_x  = np.cos(beta)*air_speed 
    v_y  = np.sin(beta)*air_speed
        
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_x 
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_y 
    segment.state.conditions.frames.inertial.acceleration_vector  = np.array([[acceleration,0.0,linear_acceleration_z]])  
    segment.state.conditions.static_stability.roll_rate[:,0]      = roll_rate         
    segment.state.conditions.static_stability.pitch_rate[:,0]     = pitch_rate
    segment.state.conditions.static_stability.yaw_rate[:,0]       = yaw_rate      

    
# ----------------------------------------------------------------------------------------------------------------------  
#  Unpack Unknowns 
# ----------------------------------------------------------------------------------------------------------------------  
def unpack_unknowns(segment):
    """ Unpacks the x accleration and body angle from the solver to the mission
    
        Assumptions:
        N/A
        
        Inputs:
            segment.state.unknowns:
                acceleration                        [meters/second^2]
                body_angle                          [radians]
            
        Outputs:
            segment.state.conditions:
                frames.inertial.acceleration_vector [meters/second^2]
                frames.body.inertial_rotations      [radians]

        Properties Used:
        N/A
                                
    """      
    
    # unpack unknowns  
    acceleration  = segment.state.unknowns.acceleration 
    segment.state.conditions.frames.inertial.acceleration_vector[0,0] = acceleration         


# RCAIDE/Methods/Energy/Propulsors/Turbofan_Propulsor/size_core.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Library.Methods.Propulsors.Turbofan_Propulsor            import compute_thrust

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  size_core
# ---------------------------------------------------------------------------------------------------------------------- 
def size_core(turbofan,turbofan_conditions,conditions):
    """Sizes the core flow for the design condition by computing the
    non-dimensional thrust 

    Assumptions:
        Working fluid is a perfect gas

    Source:
        https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/

    Args:
        conditions.freestream.speed_of_sound  (numpy.ndarray): [m/s]  
        turbofan
          .bypass_ratio                (float): bypass_ratio                [-]
          .total_temperature_reference (float): total temperature reference [K]
          .total_pressure_reference    (float): total pressure reference    [Pa]  
          .reference_temperature              (float): reference temperature       [K]
          .reference_pressure                 (float): reference pressure          [Pa]
          .design_thrust                      (float): design thrust               [N]  

    Returns:
        None 
    """             
    # Unpack flight conditions 
    a0             = conditions.freestream.speed_of_sound

    # Unpack turbofan flight conditions 
    bypass_ratio   = turbofan_conditions.bypass_ratio
    Tref           = turbofan.reference_temperature
    Pref           = turbofan.reference_pressure 
    Tt_ref         = turbofan_conditions.total_temperature_reference  
    Pt_ref         = turbofan_conditions.total_pressure_reference
    
    # Compute nondimensional thrust
    turbofan_conditions.throttle = 1.0
    compute_thrust(turbofan,turbofan_conditions,conditions) 

    # Compute dimensional mass flow rates
    TSFC       = turbofan_conditions.thrust_specific_fuel_consumption
    Fsp        = turbofan_conditions.non_dimensional_thrust
    mdot_core  = turbofan.design_thrust/(Fsp*a0*(1+bypass_ratio)*turbofan_conditions.throttle)  
    mdhc       = mdot_core/ (np.sqrt(Tref/Tt_ref)*(Pt_ref/Pref))

    # Store results on turbofan data structure 
    turbofan.TSFC                                = TSFC
    turbofan.mass_flow_rate_design               = mdot_core
    turbofan.compressor_nondimensional_massflow  = mdhc

    return  



# RCAIDE/Library/Methods/Propulsors/Converters/DC_Motor/__init__.py
# (c) Copyright 2023 Aerospace Research Community LLC

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .append_motor_conditions           import append_motor_conditions 
from .compute_motor_performance         import compute_torque_from_RPM_and_voltage
from .compute_motor_performance         import compute_RPM_and_torque_from_power_coefficent_and_voltage
from .compute_motor_performance         import compute_current_from_RPM_and_voltage
from .compute_motor_performance         import compute_voltage_and_current_from_RPM_and_speed_constant 
from .design_motor                      import design_motor



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_landing_gear_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import RCAIDE
from RCAIDE.Framework.Core import  Units , Data 
from .compute_fuselage_weight import compute_fuselage_weight
from .compute_landing_gear_weight import compute_landing_gear_weight
from .compute_payload_weight import compute_payload_weight
from .compute_systems_weight import compute_systems_weight
from .compute_horizontal_tail_weight import compute_horizontal_tail_weight
from .compute_vertical_tail_weight import compute_vertical_tail_weight
from .compute_main_wing_weight import compute_main_wing_weight
from RCAIDE.Library.Methods.Weights.Correlation_Buildups import Propulsion as Propulsion 


# ----------------------------------------------------------------------------------------------------------------------
# Main Wing Weight 
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_operating_empty_weight(vehicle, settings=None):
    """ output = RCAIDE.Methods.Weights.Correlations.Tube_Wing.empty(engine,wing,aircraft,fuselage,horizontal,vertical)
        Computes the empty weight breakdown of a General Aviation type aircraft  
        
        Inputs:
            engine - a data dictionary with the fields:                    
                thrust_sls - sea level static thrust of a single engine [Newtons]

            vehicle - a data dictionary with the fields:                    
                reference_area                                                            [meters**2]
                envelope - a data dictionary with the fields:
                    ultimate_load - ultimate load of the aircraft                         [dimensionless]
                    limit_load    - limit load factor at zero fuel weight of the aircraft [dimensionless]
                
                mass_properties - a data dictionary with the fields:
                    max_takeoff   - max takeoff weight of the vehicle           [kilograms]
                    max_zero_fuel - maximum zero fuel weight of the aircraft    [kilograms]
                    cargo         - cargo weight                                [kilograms]
                
                passengers - number of passengers on the aircraft               [dimensionless]
                        
                design_dynamic_pressure - dynamic pressure at cruise conditions [Pascal]
                design_mach_number      - mach number at cruise conditions      [dimensionless]
                
                networks - a data dictionary with the fields: 
                    keys           - identifier for the type of network; different types have different fields
                        turbofan
                            thrust_sls - sealevel standard thrust                               [Newtons]             
                        internal_combustion
                            rated_power - maximum rated power of the internal combustion engine [Watts]
                        
                    number_of_engines - integer indicating the number of engines on the aircraft

                W_cargo - weight of the bulk cargo being carried on the aircraft [kilograms]
                num_seats - number of seats installed on the aircraft [dimensionless]
                ctrl - specifies if the control system is "fully powered", "partially powered", or not powered [dimensionless]
                ac - determines type of instruments, electronics, and operating items based on types: 
                    "short-range", "medium-range", "long-range", "business", "cargo", "commuter", "sst" [dimensionless]
                w2h - tail length (distance from the airplane c.g. to the horizontal tail aerodynamic center) [meters]
                
                fuel - a data dictionary with the fields: 
                    mass_properties  - a data dictionary with the fields:
                        mass -mass of fuel [kilograms]
                    density          - gravimetric density of fuel                             [kilograms/meter**3]    
                    number_of_tanks  - number of external fuel tanks                           [dimensionless]
                    internal_volume  - internal fuel volume contained in the wing              [meters**3]
                wings - a data dictionary with the fields:    
                    wing - a data dictionary with the fields:
                        span                      - span of the wing                           [meters]
                        taper                     - taper ratio of the wing                    [dimensionless]
                        thickness_to_chord        - thickness-to-chord ratio of the wing       [dimensionless]
                        chords - a data dictionary with the fields:
                            mean_aerodynamic - mean aerodynamic chord of the wing              [meters]
                            root             - root chord of the wing                          [meters]
                            
                            
                        sweeps - a data dictionary with the fields:
                            quarter_chord - quarter chord sweep angle of the wing              [radians]
                        mac                       - mean aerodynamic chord of the wing         [meters]
                        r_c                       - wing root chord                            [meters]
                        origin  - location of the leading edge of the wing relative to the front of the fuselage                                      [meters,meters,meters]
                        aerodynamic_center - location of the aerodynamic center of the horizontal_stabilizer relative to the leading edge of the wing [meters,meters,meters]
        
                    
                    
                    
                    horizontal_stabilizer - a data dictionary with the fields:
                        areas -  a data dictionary with the fields:
                            reference - reference area of the horizontal stabilizer                                    [meters**2]
                            exposed  - exposed area for the horizontal tail                                            [meters**2]
                        taper   - taper ratio of the horizontal stabilizer                                             [dimensionless]
                        span    - span of the horizontal tail                                                          [meters]
                        sweeps - a data dictionary with the fields:
                            quarter_chord - quarter chord sweep angle of the horizontal stabilizer                     [radians]
                        chords - a data dictionary with the fields:
                            mean_aerodynamic - mean aerodynamic chord of the horizontal stabilizer                     [meters]         
                            root             - root chord of the horizontal stabilizer             
                        thickness_to_chord - thickness-to-chord ratio of the horizontal tail                           [dimensionless]
                        mac     - mean aerodynamic chord of the horizontal tail                                        [meters]
                        origin  - location of the leading of the horizontal tail relative to the front of the fuselage                                                 [meters,meters,meters]
                        aerodynamic_center - location of the aerodynamic center of the horizontal_stabilizer relative to the leading edge of the horizontal stabilizer [meters,meters,meters]
        
                    vertical - a data dictionary with the fields:
                        areas -  a data dictionary with the fields:
                            reference - reference area of the vertical stabilizer         [meters**2]
                        span    - span of the vertical tail                               [meters]
                        taper   - taper ratio of the horizontal stabilizer                [dimensionless]
                        t_c     - thickness-to-chord ratio of the vertical tail           [dimensionless]
                        sweeps   - a data dictionary with the fields:
                            quarter_chord - quarter chord sweep angle of the vertical stabilizer [radians]
                        t_tail - flag to determine if aircraft has a t-tail, "yes"               [dimensionless]


                
                fuselages - a data dictionary with the fields:  
                    fuselage - a data dictionary with the fields:
                        areas             - a data dictionary with the fields:
                            wetted - wetted area of the fuselage [meters**2]
                        differential_pressure  - Maximum fuselage pressure differential   [Pascal]
                        width             - width of the fuselage                         [meters]
                        heights - a data dictionary with the fields:
                            maximum - height of the fuselage                              [meters]
                        lengths-  a data dictionary with the fields:
                            structure - structural length of the fuselage                 [meters]                     
                        mass_properties - a data dictionary with the fields:
                            volume - total volume of the fuselage                         [meters**3]
                            internal_volume - internal volume of the fuselage             [meters**3]
                        number_coach_sets - number of seats on the aircraft               [dimensionless]    
                landing_gear - a data dictionary with the fields:
                    main - a data dictionary with the fields:
                        strut_length - strut length of the main gear                      [meters]
                    nose - a data dictionary with the fields:
                        strut_length - strut length of the nose gear                      [meters]
                avionics - a data dictionary, used to determine if avionics weight is calculated, don't include if vehicle has none
                air_conditioner - a data dictionary, used to determine if air conditioner weight is calculated, don't include if vehicle has none
        
        
        Outputs:
            output - a data dictionary with fields:
                wing - wing weight                            [kilograms]
                fuselage - fuselage weight                    [kilograms]
                propulsion - propulsion                       [kilograms]
                landing_gear_main - main gear weight          [kilograms]
                landing_gear_nose - nose gear weight          [kilograms]
                horizonal_tail - horizontal stabilizer weight [kilograms]
                vertical_tail - vertical stabilizer weight    [kilograms]
                systems - total systems weight                [kilograms]
                systems - a data dictionary with fields:
                    control_systems - control systems weight  [kilograms]
                    hydraulics - hydraulics weight            [kilograms]
                    avionics - avionics weight                [kilograms]
                    electric - electrical systems weight      [kilograms]
                    air_conditioner - air conditioner weight  [kilograms]
                    furnish - furnishing weight               [kilograms]
                    fuel_system - fuel system weight          [ kilograms]
           Wing, empannage, fuselage, propulsion and individual systems masses updated with their calculated values
       Assumptions:
            calculated aircraft weight from correlations created per component of historical aircraft
        
    """     

    if settings == None: 
        use_max_fuel_weight = True 
    else:
        use_max_fuel_weight = settings.use_max_fuel_weight
        
    # Unpack inputs
    S_gross_w   = vehicle.reference_area
    Nult        = vehicle.flight_envelope.ultimate_load 
    TOW         = vehicle.mass_properties.max_takeoff 
    num_pax     = vehicle.passengers
    W_cargo     = vehicle.mass_properties.cargo 
    q_c         = vehicle.design_dynamic_pressure
    mach_number = vehicle.flight_envelope.design_mach_number
 
    landing_weight              = TOW
    m_fuel                      =  0
    number_of_tanks             =  0
    V_fuel                      =  0
    V_fuel_int                  =  0
    W_energy_network_cumulative =  0
    number_of_engines           =  0
 
    for network in vehicle.networks:
        W_energy_network_total   = 0
        number_of_jet_engines    = 0
        number_of_piston_engines = 0  

        for fuel_line in  network.fuel_lines: 
            for fuel_tank in fuel_line.fuel_tanks: 
                m_fuel_tank     = fuel_tank.fuel.mass_properties.mass
                m_fuel          += m_fuel_tank   
                landing_weight  -= m_fuel_tank   
                number_of_tanks += 1
                V_fuel_int      += m_fuel_tank/fuel_tank.fuel.density  #assume all fuel is in integral tanks 
                V_fuel          += m_fuel_tank/fuel_tank.fuel.density #total fuel  
         
        # Electric-Powered Propulsors  
        for bus in network.busses: 
            # electrical payload 
            W_energy_network_total  += bus.payload.mass_properties.mass * Units.kg
     
            # Avionics Weight 
            W_energy_network_total  += bus.avionics.mass_properties.mass      
    
            for battery in bus.battery_modules: 
                W_energy_network_total  += battery.mass_properties.mass * Units.kg
                  
            for propulsor in bus.propulsors:
                if 'motor' in propulsor: 
                    motor_mass = propulsor.motor.mass_properties.mass       
                    W_energy_network_cumulative  += motor_mass                
        
        for propulsor in  network.propulsors: 
            if type(propulsor) ==  RCAIDE.Library.Components.Propulsors.Turbofan:
                number_of_jet_engines += 1
                thrust_sls    =  propulsor.sealevel_static_thrust 
                W_engine_jet            = Propulsion.compute_jet_engine_weight(thrust_sls)
                W_propulsion            = Propulsion.integrated_propulsion(W_engine_jet,number_of_jet_engines) 
                propulsor.mass_properties.mass = W_propulsion
                W_energy_network_total  += W_propulsion                
            elif type(propulsor) ==  RCAIDE.Library.Components.Propulsors.ICE_Propeller:      
                number_of_piston_engines += 1
                rated_power              = propulsor.engine.sea_level_power 
                W_engine_piston          = Propulsion.compute_piston_engine_weight(rated_power)
                W_propulsion             = Propulsion.integrated_propulsion_general_aviation(W_engine_piston,number_of_piston_engines) 
                propulsor.mass_properties.mass = W_propulsion
                W_energy_network_total  += W_propulsion
            elif type(propulsor) ==  RCAIDE.Library.Components.Propulsors.Turboprop:      
                number_of_piston_engines += 1
                rated_power   = propulsor.design_power 
                W_engine_piston          = Propulsion.compute_piston_engine_weight(rated_power)
                W_propulsion             = Propulsion.integrated_propulsion_general_aviation(W_engine_piston,number_of_piston_engines) 
                propulsor.mass_properties.mass = W_propulsion
                W_energy_network_total  += W_propulsion        
                
        W_energy_network_cumulative += W_energy_network_total
        number_of_engines           +=  number_of_jet_engines +  number_of_piston_engines
    
    for wing in vehicle.wings:
        if isinstance(wing,RCAIDE.Library.Components.Wings.Main_Wing):
            b          = wing.spans.projected
            AR_w       = (b**2.)/S_gross_w
            taper_w    = wing.taper
            t_c_w      = wing.thickness_to_chord
            sweep_w    = wing.sweeps.quarter_chord  
            W_wing    = compute_main_wing_weight(S_gross_w, m_fuel, AR_w, sweep_w, q_c, taper_w, t_c_w,Nult,TOW)
            wing.mass_properties.mass = W_wing
            
            # set main wing to be used in future horizontal tail calculations 
            main_wing  =  wing
        
    l_w2h = 0
    W_tail_horizontal =  0
    W_tail_vertical   =  0
    for wing in vehicle.wings:            
        if isinstance(wing,RCAIDE.Library.Components.Wings.Horizontal_Tail):
            S_h                = wing.areas.reference
            b_h                = wing.spans.projected
            AR_h               = (b_h**2.)/S_h
            taper_h            = wing.spans.projected
            sweep_h            = wing.sweeps.quarter_chord 
            t_c_h              = wing.thickness_to_chord
            l_w2h              = wing.origin[0][0] + wing.aerodynamic_center[0] - main_wing.origin[0][0] - main_wing.aerodynamic_center[0] 
            W_tail_horizontal  = compute_horizontal_tail_weight(S_h, AR_h, sweep_h, q_c, taper_h, t_c_h,Nult,TOW)                 
            wing.mass_properties.mass = W_tail_horizontal     
        if isinstance(wing,RCAIDE.Library.Components.Wings.Vertical_Tail):     
            S_v               = wing.areas.reference
            b_v               = wing.spans.projected
            AR_v              = (b_v**2.)/S_v
            taper_v           = wing.taper
            t_c_v             = wing.thickness_to_chord
            sweep_v           = wing.sweeps.quarter_chord
            t_tail            = wing.t_tail  
            W_tail_vertical   = compute_vertical_tail_weight(S_v, AR_v, sweep_v, q_c, taper_v, t_c_v, Nult,TOW,t_tail) 
            wing.mass_properties.mass = W_tail_vertical
    
    for fuselage in  vehicle.fuselages: 
        S_fus       = fuselage.areas.wetted
        diff_p_fus  = fuselage.differential_pressure
        w_fus       = fuselage.width
        h_fus       = fuselage.heights.maximum 
        l_fus       = fuselage.lengths.total-fuselage.lengths.tail  
        V_fuse      = fuselage.mass_properties.volume 
        num_seats   = fuselage.number_coach_seats  
        W_fuselage  = compute_fuselage_weight(S_fus, Nult, TOW, w_fus, h_fus, l_fus, l_w2h, q_c, V_fuse, diff_p_fus)
        fuselage.mass_properties.mass = W_fuselage
        
    # landing gear 
    strut_length_main = 0
    strut_length_nose = 0 
    nose_landing_gear = False
    main_landing_gear = False
    for LG in vehicle.landing_gears:
        if isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            strut_length_main = LG.strut_length
            main_landing_gear = True
        elif isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear):
            strut_length_nose = LG.strut_length 
            nose_landing_gear = True
    W_landing_gear         = compute_landing_gear_weight(landing_weight, Nult, strut_length_main, strut_length_nose) 
    for landing_gear in vehicle.landing_gears:
        if isinstance(landing_gear, RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            landing_gear.mass_properties.mass = W_landing_gear.main
            main_landing_gear = True
        elif isinstance(landing_gear, RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear):
            landing_gear.mass_properties.mass = W_landing_gear.nose
            nose_landing_gear = True 
    if nose_landing_gear == False:
        nose_gear = RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear()  
        nose_gear.mass_properties.mass = W_landing_gear.nose
        vehicle.append_component(nose_gear) 
    if main_landing_gear == False:
        main_gear = RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear()  
        main_gear.mass_properties.mass = W_landing_gear.main
        vehicle.append_component(main_gear)
         
    if len(vehicle.avionics) == 0:
        avionics     = RCAIDE.Library.Components.Systems.Avionics()
        W_uav        = 0. 
    else:
        avionics = vehicle.avionics
        W_uav    = avionics.mass_properties.uninstalled

    has_air_conditioner = 0
    if 'air_conditioner' in vehicle:
        has_air_conditioner = 1

    # Calculating Empty Weight of Aircraft
    W_systems           = compute_systems_weight(W_uav,V_fuel, V_fuel_int, number_of_tanks, number_of_engines, l_fus, b, TOW, Nult, num_seats, mach_number, has_air_conditioner)

    # Calculate the equipment empty weight of the aircraft

    W_empty           = (W_wing + W_fuselage + W_landing_gear.main+W_landing_gear.nose + W_energy_network_cumulative + W_systems.total + \
                          W_tail_horizontal +W_tail_vertical) 

    # packup outputs
    W_payload = compute_payload_weight(TOW, W_empty, num_pax,W_cargo)
    
    vehicle.payload.passengers = RCAIDE.Library.Components.Component()
    vehicle.payload.baggage    = RCAIDE.Library.Components.Component()
    vehicle.payload.cargo      = RCAIDE.Library.Components.Component()
    
    vehicle.payload.passengers.mass_properties.mass = W_payload.passengers
    vehicle.payload.baggage.mass_properties.mass    = W_payload.baggage
    vehicle.payload.cargo.mass_properties.mass      = W_payload.cargo        


    # Distribute all weight in the output fields
    output                                    = Data()
    output.empty                              = Data()
    output.empty.structural                   = Data()
    output.empty.structural.wings             = W_wing +  W_tail_horizontal + W_tail_vertical 
    output.empty.structural.fuselage          = W_fuselage
    output.empty.structural.landing_gear      = W_landing_gear.main +  W_landing_gear.nose 
    output.empty.structural.nacelle           = 0
    output.empty.structural.paint             = 0  
    output.empty.structural.total             = output.empty.structural.wings \
                                                     + output.empty.structural.fuselage  + output.empty.structural.landing_gear \
                                                     + output.empty.structural.paint + output.empty.structural.nacelle
          
    output.empty.propulsion                   = Data()
    output.empty.propulsion.total             = W_energy_network_cumulative
    output.empty.propulsion.fuel_system       = W_systems.W_fuel_system
  
    output.empty.systems                      = Data()
    output.empty.systems.control_systems      = W_systems.W_flight_control
    output.empty.systems.hydraulics           = W_systems.W_hyd_pnu
    output.empty.systems.avionics             = W_systems.W_avionics
    output.empty.systems.electrical           = W_systems.W_electrical
    output.empty.systems.air_conditioner      = W_systems.W_ac
    output.empty.systems.furnishings              = W_systems.W_furnish
    output.empty.systems.apu                  = 0
    output.empty.systems.instruments          = 0
    output.empty.systems.anti_ice             = 0
    output.empty.systems.total                = output.empty.systems.control_systems + output.empty.systems.apu \
                                                  + output.empty.systems.electrical + output.empty.systems.avionics \
                                                  + output.empty.systems.hydraulics + output.empty.systems.furnishings \
                                                  + output.empty.systems.air_conditioner + output.empty.systems.instruments \
                                                  + output.empty.systems.anti_ice
  
    output.payload                                = Data()
    output.payload                                = W_payload
    output.operational_items                      = Data()
    output.operational_items.oper_items           = 0
    output.operational_items.flight_crew          = 0
    output.operational_items.flight_attendants    = 0
    output.operational_items.total                = 0

    output.empty.total      = output.empty.structural.total + output.empty.propulsion.total + output.empty.systems.total
    output.operating_empty  = output.empty.total + output.operational_items.total
    output.zero_fuel_weight =  output.operating_empty + output.payload.total 

    if use_max_fuel_weight:  # assume fuel is equally distributed in fuel tanks
        total_fuel_weight  = vehicle.mass_properties.max_takeoff -  output.zero_fuel_weight
        for network in vehicle.networks: 
            for fuel_line in network.fuel_lines:  
                for fuel_tank in fuel_line.fuel_tanks:
                    fuel_weight =  total_fuel_weight/number_of_tanks  
                    fuel_tank.fuel.mass_properties.mass = fuel_weight
        output.fuel = total_fuel_weight 
        output.total = output.zero_fuel_weight + output.fuel
    else:
        total_fuel_weight =  0
        for network in vehicle.networks: 
            for fuel_line in network.fuel_lines:  
                for fuel_tank in fuel_line.fuel_tanks:
                    fuel_mass =  fuel_tank.fuel.density * fuel_tank.volume
                    fuel_tank.fuel.mass_properties.mass = fuel_mass * 9.81
                    total_fuel_weight = fuel_mass * 9.81 
        output.fuel = total_fuel_weight
        output.total = output.zero_fuel_weight + output.fuel  
      
    control_systems                                  = RCAIDE.Library.Components.Component()
    control_systems.tag                              = 'control_systems'  
    electrical_systems                               = RCAIDE.Library.Components.Component()
    electrical_systems.tag                           = 'electrical_systems'
    furnishings                                      = RCAIDE.Library.Components.Component()
    furnishings.tag                                  = 'furnishings'
    air_conditioner                                  = RCAIDE.Library.Components.Component() 
    air_conditioner.tag                              = 'air_conditioner' 
    hydraulics                                       = RCAIDE.Library.Components.Component()
    hydraulics.tag                                   = 'hydraulics'  
     
    nose_landing_gear = False
    main_landing_gear =  False
    for LG in vehicle.landing_gears:
        if isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear):
            LG.mass_properties.mass = W_landing_gear.main
            main_landing_gear = True
        elif isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear):
            LG.mass_properties.mass = W_landing_gear.nose
            nose_landing_gear = True 
    if nose_landing_gear == False:
        nose_gear = RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear()  
        nose_gear.mass_properties.mass = W_landing_gear.nose 
        vehicle.append_component(nose_gear) 
    if main_landing_gear == False:
        main_gear = RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear()  
        main_gear.mass_properties.mass = W_landing_gear.main
        vehicle.append_component(main_gear)
         
    control_systems.mass_properties.mass    = output.empty.systems.control_systems
    electrical_systems.mass_properties.mass = output.empty.systems.electrical
    furnishings.mass_properties.mass        = output.empty.systems.furnishings
    avionics.mass_properties.mass           = output.empty.systems.avionics + output.empty.systems.instruments
    air_conditioner.mass_properties.mass    = output.empty.systems.air_conditioner 
    hydraulics.mass_properties.mass         = output.empty.systems.hydraulics

    # assign components to vehicle
    vehicle.control_systems                             = control_systems
    vehicle.electrical_systems                          = electrical_systems
    vehicle.avionics                                    = avionics
    vehicle.furnishings                                 = furnishings 
    vehicle.hydraulics                                  = hydraulics
    if has_air_conditioner:
        vehicle.air_conditioner.mass_properties.mass    = output.empty.systems.air_conditioner 
    return output


# RCAIDE/Methods/Aerodynamics/Common/Drag/parasite_drag_total.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
import RCAIDE
import numpy as np 
 
# package imports
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------  
#  Total Parasite Drag 
# ----------------------------------------------------------------------------------------------------------------------   
def parasite_total(state,settings,geometry):
    """Sums up the parasite drags from all compoments 

    Assumptions:
        None

    Source:
        None

    Args:
        state                                                (dict): flight conditions       [-]
        settings                                             (dict): analyses settings       [-]
        geometry                                             (dict): aircraft data stucture  [-]
        conditions.aerodynamics.coefficients.drag.
          parasite[wing.tag(s)].parasite_drag       (numpy.ndarray): wing parasite drags     [Unitless]
          parasite[fuselage.tag(s)].parasite_drag   (numpy.ndarray): fuselage parasite drags [Unitless]
          parasite[boom.tag(s)].parasite_drag       (numpy.ndarray): boom parasite drags     [Unitless]
          parasite[nacelle.tag(s)].parasite_drag    (numpy.ndarray): nacelle parasite drags  [Unitless]
          parasite[pylon.tag(s)].parasite_drag      (numpy.ndarray): pylon parasite drags    [Unitless]


    Returns:
        None 
    """ 
    # unpack
    conditions             = state.conditions 
    vehicle_reference_area = geometry.reference_area
    
    #compute parasite drag total
    total_parasite_drag = 0.0
    
    # renormalize parasite drag from wings using reference area of aircraft 
    for wing in geometry.wings:
        parasite_drag = conditions.aerodynamics.coefficients.drag.parasite[wing.tag].total 
        conditions.aerodynamics.coefficients.drag.parasite[wing.tag].total  = parasite_drag * wing.areas.reference/vehicle_reference_area
        total_parasite_drag += parasite_drag * wing.areas.reference/vehicle_reference_area
 
    # renormalize parasite drag from fuselages using reference area of aircraft 
    for fuselage in geometry.fuselages:
        if type(fuselage) == RCAIDE.Library.Components.Fuselages.Blended_Wing_Body_Fuselage:
            continue
        parasite_drag = conditions.aerodynamics.coefficients.drag.parasite[fuselage.tag].total 
        conditions.aerodynamics.coefficients.drag.parasite[fuselage.tag].total = parasite_drag * fuselage.areas.front_projected/vehicle_reference_area
        total_parasite_drag += parasite_drag * fuselage.areas.front_projected/vehicle_reference_area
    
    # renormalize parasite drag from nacelles and pylons using reference area of aircraft  
    for network in  geometry.networks: 
        for propulsor in network.propulsors:  
            if 'nacelle' in propulsor:
                nacelle       =  propulsor.nacelle 
                ref_area      = nacelle.diameter**2 / 4 * np.pi
                parasite_drag = conditions.aerodynamics.coefficients.drag.parasite[nacelle.tag].total 
                conditions.aerodynamics.coefficients.drag.parasite[nacelle.tag].total  = parasite_drag * ref_area/vehicle_reference_area
                total_parasite_drag += parasite_drag * ref_area/vehicle_reference_area 
                    
                if nacelle.has_pylon:
                    parasite_drag = conditions.aerodynamics.coefficients.drag.parasite[nacelle.tag + '_pylon'].total 
                    conditions.aerodynamics.coefficients.drag.parasite[nacelle.tag + '_pylon'].total = parasite_drag * fuselage.areas.front_projected/vehicle_reference_area
                    total_parasite_drag += parasite_drag * ref_area/vehicle_reference_area 
    
    state.conditions.aerodynamics.coefficients.drag.parasite.total = total_parasite_drag

    return 


## @ingroup Optimization-Package_Setups-TRMM
# TRMM_setup.py
#
# Created:  Apr 2017, T. MacDonald
# Modified: 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

from RCAIDE.Framework.Optimization.Packages.trmm import Trust_Region_Optimization as tro
from RCAIDE.Framework.Optimization.Packages.trmm.Trust_Region import Trust_Region

# ----------------------------------------------------------------------
#  TRMM_Solve
# ----------------------------------------------------------------------

## @ingroup Optimization-Package_Setups-TRMM
def TRMM_Solve(problem,tr=None,tr_opt=None,print_output=False):
    """ This solves your TRMM

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        problem       [nexus()]
        tr            [Trust_Region()]
        tr_opt        [Trust_Region_Optimization()]
        print_output  [bool]

        Outputs:
        None

        Properties Used:
        None
        """     

    if tr == None:
        tr = Trust_Region()
    problem.trust_region = tr
    if tr_opt == None:
        TRM_opt = tro.Trust_Region_Optimization()
    else:
        TRM_opt = tr_opt
    TRM_opt.optimize(problem,print_output=print_output)

    return





# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# RCAIDE Imports
from  RCAIDE import * 
from   RCAIDE                    import * 
from   RCAIDE.Library.Components import Wings
from   RCAIDE.Framework.Core     import Units, Data

# ----------------------------------------------------------------------
#  Compute drag of turbofan in windmilling condition
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Helper_Functions
def windmilling_drag(geometry,state):
    """Computes windmilling drag for turbofan engines

    Assumptions:
    None

    Source:
    http://www.dept.aoe.vt.edu/~mason/Mason_f/AskinThesis2002_13.pdf

    Inputs:
    geometry.
      max_mach_operational        [Unitless]
      reference_area              [m^2]
      wings.sref                  [m^2]
      networks.
        areas.wetted              [m^2]
        length                    [m]

    Outputs:
    windmilling_drag_coefficient  [Unitless]

    Properties Used:
    N/A
    """    
    # ==============================================
        # Unpack
    # ==============================================
    vehicle = geometry

    # Defining reference area
    if vehicle.reference_area:
        reference_area = vehicle.reference_area
    else:
        n_wing = 0
        for wing in vehicle.wings:
            if not isinstance(wing,Wings.Main_Wing): continue
            n_wing = n_wing + 1
            reference_area = wing.sref
        if n_wing > 1:
            print(' More than one Main_Wing in the vehicle. Last one will be considered.')
        elif n_wing == 0:
            print('No Main_Wing defined! Using the 1st wing found')
            for wing in vehicle.wings:
                if not isinstance(wing,Wings.Wing): continue
                reference_area = wing.sref
                break

    # getting geometric data from engine (estimating when not available)
    swet_nac = 0

    for network in vehicle.networks: 
        for propulsor in network.propulsors:  
            if 'nacelle' in propulsor:
                swet_nac += propulsor.nacelle.areas.wetted 
    # Compute
    windmilling_drag_coefficient = 0.007274 * swet_nac / reference_area

    # dump data to state
    windmilling_result = Data(
        wetted_area  = swet_nac    ,
        total        = windmilling_drag_coefficient ,
    )
    state.conditions.aerodynamics.coefficients.drag.windmilling = windmilling_result

    return windmilling_drag_coefficient


## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
# deflect_control_surface.py
# 
# Created:  Jul 2022, A. Blaufox & E. Botero
# Modified:
#           

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np
from RCAIDE.Library.Components.Wings import All_Moving_Surface
from RCAIDE.Framework.Core import Data
from .generate_VD_helpers import postprocess_VD

# ----------------------------------------------------------------------
#  Deflect Control Surface
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
def deflect_control_surfaces(VD,geometry,settings):
    """ 
    Goes through a vehicle and updates the control surface deflections in the VD. Crucially this rebuilds the VD as a
    postprocess step
    
    Assumptions: 

    Source:  


    Inputs: 
    VD - vehicle vortex distribution              [Unitless] 
    geometry.wings                                [Unitless]  
    settings.floating_point_precision             [np.dtype]

    Outputs:      
    VD - vehicle vortex distribution              [Unitless] 

    Properties Used:
    N/A
    """     
    
    # Loop over t ewings
    for wing in VD.VLM_wings:
        wing_is_all_moving = (not wing.is_a_control_surface) and issubclass(wing.wing_type, All_Moving_Surface)        
        if wing.is_a_control_surface or wing_is_all_moving:
            # Deflect the control surface
            VD, wing = deflect_control_surface(VD, wing)    
        
    VD = postprocess_VD(VD, settings)
    
    # pack VD into geometry
    geometry.vortex_distribution = VD    
    
    
    return VD


## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
def deflect_control_surface(VD,wing):
    """ 
    Deflects the panels of a vortex distribution that correspond to the given VLM_wing. 
    
    Assumptions: 
    If the user calls this function outside of generate_vortex_distribution,
    RCAIDE.Methods.Aerodynamics.Common.Fidelity_Zero.Lift.postprocess_VD MUST be called
    right after

    Source:  


    Inputs: 
    VD                   - vehicle vortex distribution                    [Unitless] 
    wing                 - a VLM_wing object that was generated in the    [Unitless] 
                           original generate_vortex_distribution call. 
    wing.deflection_last - last deflection applied to this wing           [radians] 
    wing.deflection      - deflection to set this wing to.                [radians] 
    
    Outputs:      
    VD       - vehicle vortex distribution                    [Unitless] 
    wing     - VLM_wing object                                [Unitless] 


    Properties Used:
    N/A
    """     
    
    # Unpack number of strips for this wing
    n_sw     = wing.n_sw
    n_cw     = wing.n_cw
    sym_para = wing.symmetric    


    # Symmetry loop
    signs         = np.array([1, -1], dtype=int) # acts as a multiplier for symmetry. -1 is only ever used for symmetric wings
    symmetry_mask = [True,sym_para]
    for sym_sign in signs[symmetry_mask]:    
        
        # Pull out initial VD data points of surface
        condition      = VD.surface_ID      == wing.surface_ID*sym_sign
        condition_full = VD.surface_ID_full == wing.surface_ID*sym_sign
        xi_prime_a1    = VD.XA1[condition]
        xi_prime_ac    = VD.XAC[condition]
        xi_prime_ah    = VD.XAH[condition]
        xi_prime_a2    = VD.XA2[condition]
        y_prime_a1     = VD.YA1[condition]
        y_prime_ah     = VD.YAH[condition]
        y_prime_ac     = VD.YAC[condition]
        y_prime_a2     = VD.YA2[condition]
        zeta_prime_a1  = VD.ZA1[condition]
        zeta_prime_ah  = VD.ZAH[condition]
        zeta_prime_ac  = VD.ZAC[condition]
        zeta_prime_a2  = VD.ZA2[condition]
        xi_prime_b1    = VD.XB1[condition]
        xi_prime_bh    = VD.XBH[condition]
        xi_prime_bc    = VD.XBC[condition]
        xi_prime_b2    = VD.XB2[condition]
        y_prime_b1     = VD.YB1[condition]
        y_prime_bh     = VD.YBH[condition]
        y_prime_bc     = VD.YBC[condition]
        y_prime_b2     = VD.YB2[condition]
        zeta_prime_b1  = VD.ZB1[condition]
        zeta_prime_bh  = VD.ZBH[condition]
        zeta_prime_bc  = VD.ZBC[condition]
        zeta_prime_b2  = VD.ZB2[condition]
        xi_prime_ch    = VD.XCH[condition]
        xi_prime       = VD.XC [condition]
        y_prime_ch     = VD.YCH[condition]
        y_prime        = VD.YC [condition]
        zeta_prime_ch  = VD.ZCH[condition]
        zeta_prime     = VD.ZC [condition]
        
        X_as = np.zeros_like(VD.X[condition_full][:-(n_cw+1)])
        Y_as = np.zeros_like(VD.Y[condition_full][:-(n_cw+1)])
        Z_as = np.zeros_like(VD.Z[condition_full][:-(n_cw+1)])
        
        
        for idx_y in range(n_sw):
            start     , stop      = idx_y*n_cw    , (idx_y+1)*n_cw
            start_full, stop_full = idx_y*(n_cw+1), (idx_y+1)*(n_cw+1)
            
            # pack strip values
            raw_VD = Data()
            raw_VD.xi_prime_a1   = xi_prime_a1  [start:stop]
            raw_VD.xi_prime_ac   = xi_prime_ac  [start:stop]
            raw_VD.xi_prime_ah   = xi_prime_ah  [start:stop]
            raw_VD.xi_prime_a2   = xi_prime_a2  [start:stop]
            raw_VD.y_prime_a1    = y_prime_a1   [start:stop]
            raw_VD.y_prime_ah    = y_prime_ah   [start:stop]
            raw_VD.y_prime_ac    = y_prime_ac   [start:stop]
            raw_VD.y_prime_a2    = y_prime_a2   [start:stop]
            raw_VD.zeta_prime_a1 = zeta_prime_a1[start:stop]
            raw_VD.zeta_prime_ah = zeta_prime_ah[start:stop]
            raw_VD.zeta_prime_ac = zeta_prime_ac[start:stop]
            raw_VD.zeta_prime_a2 = zeta_prime_a2[start:stop]
            raw_VD.xi_prime_b1   = xi_prime_b1  [start:stop]
            raw_VD.xi_prime_bh   = xi_prime_bh  [start:stop]
            raw_VD.xi_prime_bc   = xi_prime_bc  [start:stop]
            raw_VD.xi_prime_b2   = xi_prime_b2  [start:stop]
            raw_VD.y_prime_b1    = y_prime_b1   [start:stop]
            raw_VD.y_prime_bh    = y_prime_bh   [start:stop]
            raw_VD.y_prime_bc    = y_prime_bc   [start:stop]
            raw_VD.y_prime_b2    = y_prime_b2   [start:stop]
            raw_VD.zeta_prime_b1 = zeta_prime_b1[start:stop]
            raw_VD.zeta_prime_bh = zeta_prime_bh[start:stop]
            raw_VD.zeta_prime_bc = zeta_prime_bc[start:stop]
            raw_VD.zeta_prime_b2 = zeta_prime_b2[start:stop]
            raw_VD.xi_prime_ch   = xi_prime_ch  [start:stop]
            raw_VD.xi_prime      = xi_prime     [start:stop]
            raw_VD.y_prime_ch    = y_prime_ch   [start:stop]
            raw_VD.y_prime       = y_prime      [start:stop]
            raw_VD.zeta_prime_ch = zeta_prime_ch[start:stop]
            raw_VD.zeta_prime    = zeta_prime   [start:stop]
            
            # deflect the surface
            raw_VD = deflect_control_surface_strip(wing, raw_VD, idx_y==0, sym_sign)
            
            # unpack strip values into surface values
            xi_prime_a1  [start:stop]  = raw_VD.xi_prime_a1  
            xi_prime_ac  [start:stop]  = raw_VD.xi_prime_ac  
            xi_prime_ah  [start:stop]  = raw_VD.xi_prime_ah  
            xi_prime_a2  [start:stop]  = raw_VD.xi_prime_a2  
            y_prime_a1   [start:stop]  = raw_VD.y_prime_a1   
            y_prime_ah   [start:stop]  = raw_VD.y_prime_ah   
            y_prime_ac   [start:stop]  = raw_VD.y_prime_ac   
            y_prime_a2   [start:stop]  = raw_VD.y_prime_a2   
            zeta_prime_a1[start:stop]  = raw_VD.zeta_prime_a1
            zeta_prime_ah[start:stop]  = raw_VD.zeta_prime_ah
            zeta_prime_ac[start:stop]  = raw_VD.zeta_prime_ac
            zeta_prime_a2[start:stop]  = raw_VD.zeta_prime_a2
        
            xi_prime_b1  [start:stop]  = raw_VD.xi_prime_b1  
            xi_prime_bh  [start:stop]  = raw_VD.xi_prime_bh  
            xi_prime_bc  [start:stop]  = raw_VD.xi_prime_bc  
            xi_prime_b2  [start:stop]  = raw_VD.xi_prime_b2  
            y_prime_b1   [start:stop]  = raw_VD.y_prime_b1   
            y_prime_bh   [start:stop]  = raw_VD.y_prime_bh   
            y_prime_bc   [start:stop]  = raw_VD.y_prime_bc   
            y_prime_b2   [start:stop]  = raw_VD.y_prime_b2   
            zeta_prime_b1[start:stop]  = raw_VD.zeta_prime_b1
            zeta_prime_bh[start:stop]  = raw_VD.zeta_prime_bh
            zeta_prime_bc[start:stop]  = raw_VD.zeta_prime_bc
            zeta_prime_b2[start:stop]  = raw_VD.zeta_prime_b2
               
            xi_prime_ch  [start:stop]  = raw_VD.xi_prime_ch  
            xi_prime     [start:stop]  = raw_VD.xi_prime     
            y_prime_ch   [start:stop]  = raw_VD.y_prime_ch   
            y_prime      [start:stop]  = raw_VD.y_prime      
            zeta_prime_ch[start:stop]  = raw_VD.zeta_prime_ch
            zeta_prime   [start:stop]  = raw_VD.zeta_prime   
            
            X_as[start_full:stop_full] = np.append(raw_VD.xi_prime_a1  , raw_VD.xi_prime_a2  [-1])
            Y_as[start_full:stop_full] = np.append(raw_VD.y_prime_a1   , raw_VD.y_prime_a2   [-1])
            Z_as[start_full:stop_full] = np.append(raw_VD.zeta_prime_a1, raw_VD.zeta_prime_a2[-1])
        
        # pack surface VD values into vehicle VD    
        VD.XA1[condition]    = xi_prime_a1    
        VD.XAC[condition]    = xi_prime_ac    
        VD.XAH[condition]    = xi_prime_ah    
        VD.XA2[condition]    = xi_prime_a2    
        VD.YA1[condition]    = y_prime_a1     
        VD.YAH[condition]    = y_prime_ah     
        VD.YAC[condition]    = y_prime_ac     
        VD.YA2[condition]    = y_prime_a2     
        VD.ZA1[condition]    = zeta_prime_a1  
        VD.ZAH[condition]    = zeta_prime_ah  
        VD.ZAC[condition]    = zeta_prime_ac  
        VD.ZA2[condition]    = zeta_prime_a2  
        VD.XB1[condition]    = xi_prime_b1    
        VD.XBH[condition]    = xi_prime_bh    
        VD.XBC[condition]    = xi_prime_bc    
        VD.XB2[condition]    = xi_prime_b2    
        VD.YB1[condition]    = y_prime_b1     
        VD.YBH[condition]    = y_prime_bh     
        VD.YBC[condition]    = y_prime_bc     
        VD.YB2[condition]    = y_prime_b2     
        VD.ZB1[condition]    = zeta_prime_b1  
        VD.ZBH[condition]    = zeta_prime_bh  
        VD.ZBC[condition]    = zeta_prime_bc  
        VD.ZB2[condition]    = zeta_prime_b2  
        VD.XCH[condition]    = xi_prime_ch    
        VD.XC [condition]    = xi_prime       
        VD.YCH[condition]    = y_prime_ch     
        VD.YC [condition]    = y_prime        
        VD.ZCH[condition]    = zeta_prime_ch  
        VD.ZC [condition]    = zeta_prime    
        
        X_last_bs = np.append(raw_VD.xi_prime_b1  , raw_VD.xi_prime_b2  [-1])
        Y_last_bs = np.append(raw_VD.y_prime_b1   , raw_VD.y_prime_b2   [-1])
        Z_last_bs = np.append(raw_VD.zeta_prime_b1, raw_VD.zeta_prime_b2[-1])
        
        VD.X[condition_full] = np.append(X_as, X_last_bs)
        VD.Y[condition_full] = np.append(Y_as, Y_last_bs)
        VD.Z[condition_full] = np.append(Z_as, Z_last_bs)
        
        
    wing.deflection_last = wing.deflection*1.
    
    VD.is_postprocessed = False
    
    return VD, wing



# ----------------------------------------------------------------------
#  Deflect Control Surface Strip
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
def deflect_control_surface_strip(wing, raw_VD, is_first_strip, sym_sign):
    """ Rotates existing points in the VD with respect to current values of a delta deflection

    Assumptions: 

    Source:  


    Inputs: 
    wing                 - a VLM_wing object that was generated in the    [Unitless] 
                           original generate_vortex_distribution call. 
    wing.deflection_last - last deflection applied to this wing           [radians]
    wing.deflection      - deflection to set this wing to.                [radians]
    
    raw_VD               - undeflected VD pertaining a strip of wing      [Unitless] 
    is_first_strip       - whether this is the first strip of wing        [Unitless]
    sym_sign             - 1 for original side, -1 for symmetric side     [Unitless]
    
    Outputs:      
    raw_VD - deflected VD values pertaining to wing         [Unitless] 


    Properties Used:
    N/A
    """    
    
    # Unpack
    vertical_wing = wing.vertical
    inverted_wing = wing.inverted_wing
    
    x_origin, y_origin, z_origin = wing.origin[0]
    
    y_origin *= sym_sign
    
    xi_prime_a1    = raw_VD.xi_prime_a1   
    xi_prime_ac    = raw_VD.xi_prime_ac   
    xi_prime_ah    = raw_VD.xi_prime_ah   
    xi_prime_a2    = raw_VD.xi_prime_a2   
    y_prime_a1     = raw_VD.y_prime_a1    
    y_prime_ah     = raw_VD.y_prime_ah    
    y_prime_ac     = raw_VD.y_prime_ac    
    y_prime_a2     = raw_VD.y_prime_a2    
    zeta_prime_a1  = raw_VD.zeta_prime_a1 
    zeta_prime_ah  = raw_VD.zeta_prime_ah 
    zeta_prime_ac  = raw_VD.zeta_prime_ac 
    zeta_prime_a2  = raw_VD.zeta_prime_a2 
                                          
    xi_prime_b1    = raw_VD.xi_prime_b1   
    xi_prime_bh    = raw_VD.xi_prime_bh   
    xi_prime_bc    = raw_VD.xi_prime_bc   
    xi_prime_b2    = raw_VD.xi_prime_b2   
    y_prime_b1     = raw_VD.y_prime_b1    
    y_prime_bh     = raw_VD.y_prime_bh    
    y_prime_bc     = raw_VD.y_prime_bc    
    y_prime_b2     = raw_VD.y_prime_b2    
    zeta_prime_b1  = raw_VD.zeta_prime_b1 
    zeta_prime_bh  = raw_VD.zeta_prime_bh 
    zeta_prime_bc  = raw_VD.zeta_prime_bc 
    zeta_prime_b2  = raw_VD.zeta_prime_b2 
                                          
    xi_prime_ch    = raw_VD.xi_prime_ch   
    xi_prime       = raw_VD.xi_prime      
    y_prime_ch     = raw_VD.y_prime_ch    
    y_prime        = raw_VD.y_prime       
    zeta_prime_ch  = raw_VD.zeta_prime_ch 
    zeta_prime     = raw_VD.zeta_prime    

    # flip over y = z for a vertical wing since deflection math assumes horizontal wing--------------------
    y_prime_a1, zeta_prime_a1 = flip_1(y_prime_a1, zeta_prime_a1, vertical_wing, inverted_wing)
    y_prime_ah, zeta_prime_ah = flip_1(y_prime_ah, zeta_prime_ah, vertical_wing, inverted_wing)
    y_prime_ac, zeta_prime_ac = flip_1(y_prime_ac, zeta_prime_ac, vertical_wing, inverted_wing)
    y_prime_a2, zeta_prime_a2 = flip_1(y_prime_a2, zeta_prime_a2, vertical_wing, inverted_wing)
    y_prime_b1, zeta_prime_b1 = flip_1(y_prime_b1, zeta_prime_b1, vertical_wing, inverted_wing)
    y_prime_bh, zeta_prime_bh = flip_1(y_prime_bh, zeta_prime_bh, vertical_wing, inverted_wing)
    y_prime_bc, zeta_prime_bc = flip_1(y_prime_bc, zeta_prime_bc, vertical_wing, inverted_wing)
    y_prime_b2, zeta_prime_b2 = flip_1(y_prime_b2, zeta_prime_b2, vertical_wing, inverted_wing)
    y_prime_ch, zeta_prime_ch = flip_1(y_prime_ch, zeta_prime_ch, vertical_wing, inverted_wing)
    y_prime   , zeta_prime    = flip_1(y_prime   , zeta_prime   , vertical_wing, inverted_wing)

    # Deflect control surfaces-----------------------------------------------------------------------------
    # note:    "positve" deflection corresponds to the RH rule where the axis of rotation is the OUTBOARD-pointing hinge vector
    # symmetry: the LH rule is applied to the reflected surface for non-ailerons. Ailerons follow a RH rule for both sides
    wing_is_all_moving = (not wing.is_a_control_surface) and issubclass(wing.wing_type, All_Moving_Surface)
        
    #For the first strip of the wing, always need to find the hinge root point. The hinge root point and direction vector 
    #found here will not change for the rest of this control surface/all-moving surface. See docstring for reasoning.
    if is_first_strip:
        # get rotation points by iterpolating between strip corners --> le/te, ib/ob = leading/trailing edge, in/outboard
        ib_le_strip_corner = np.array([xi_prime_a1[0 ], y_prime_a1[0 ], zeta_prime_a1[0 ]]) 
        ib_te_strip_corner = np.array([xi_prime_a2[-1], y_prime_a2[-1], zeta_prime_a2[-1]])                    
        
        interp_fractions   = np.array([0.,    2.,    4.   ]) + wing.hinge_fraction
        interp_domains     = np.array([0.,1., 2.,3., 4.,5.])
        interp_ranges_ib   = np.array([ib_le_strip_corner, ib_te_strip_corner]).T.flatten()
        ib_hinge_point     = np.interp(interp_fractions, interp_domains, interp_ranges_ib)
        
        #Find the hinge_vector if this is a control surface or the user has not already defined and chosen to use a specific one                    
        if wing.is_a_control_surface:
            need_to_compute_hinge_vector = True
        else: #wing is an all-moving surface
            hinge_vector                 = wing.hinge_vector
            hinge_vector_is_pre_defined  = (not wing.use_constant_hinge_fraction) and \
                                            not (hinge_vector==np.array([0.,0.,0.])).all()
            need_to_compute_hinge_vector = not hinge_vector_is_pre_defined  
            
        if need_to_compute_hinge_vector:
            ob_le_strip_corner = np.array([xi_prime_b1[0 ], y_prime_b1[0 ], zeta_prime_b1[0 ]])                
            ob_te_strip_corner = np.array([xi_prime_b2[-1], y_prime_b2[-1], zeta_prime_b2[-1]])                         
            interp_ranges_ob   = np.array([ob_le_strip_corner, ob_te_strip_corner]).T.flatten()
            ob_hinge_point     = np.interp(interp_fractions, interp_domains, interp_ranges_ob)
        
            use_root_chord_in_plane_normal = wing_is_all_moving and not wing.use_constant_hinge_fraction
            if use_root_chord_in_plane_normal: ob_hinge_point[0] = ib_hinge_point[0]
        
            hinge_vector       = ob_hinge_point - ib_hinge_point
            hinge_vector       = hinge_vector / np.linalg.norm(hinge_vector)   
        elif wing.vertical: #For a vertical all-moving surface, flip y and z of hinge vector before flipping again later
            hinge_vector[1], hinge_vector[2] = hinge_vector[2], hinge_vector[1] 
            
        #store hinge root point and direction vector
        wing.hinge_root_point = ib_hinge_point
        wing.hinge_vector     = hinge_vector 
        #END first strip calculations
    
    # get deflection angle
    ddeflection      = wing.deflection      - wing.deflection_last               # This is a delta deflection
    slat_multiplier  = (1 - wing.is_slat)   - wing.is_slat                       # Flip signs if it's a slat
    sym_multiplier   = (1 - (sym_sign==-1)) - wing.sign_duplicate*(sym_sign==-1) # If it's the symmetric side
    ver_multiplier   = (1 - vertical_wing)-1*vertical_wing                       # Vertical multiplier

    delta_deflection = slat_multiplier*sym_multiplier*ver_multiplier*ddeflection
        
    # make quaternion rotation matrix
    quaternion   = make_hinge_quaternion(wing.hinge_root_point, wing.hinge_vector, delta_deflection)
    
    # rotate strips
    xi_prime_a1, y_prime_a1, zeta_prime_a1 = rotate_points_with_quaternion(quaternion, [xi_prime_a1,y_prime_a1,zeta_prime_a1])
    xi_prime_ah, y_prime_ah, zeta_prime_ah = rotate_points_with_quaternion(quaternion, [xi_prime_ah,y_prime_ah,zeta_prime_ah])
    xi_prime_ac, y_prime_ac, zeta_prime_ac = rotate_points_with_quaternion(quaternion, [xi_prime_ac,y_prime_ac,zeta_prime_ac])
    xi_prime_a2, y_prime_a2, zeta_prime_a2 = rotate_points_with_quaternion(quaternion, [xi_prime_a2,y_prime_a2,zeta_prime_a2])
                                                                                                                             
    xi_prime_b1, y_prime_b1, zeta_prime_b1 = rotate_points_with_quaternion(quaternion, [xi_prime_b1,y_prime_b1,zeta_prime_b1])
    xi_prime_bh, y_prime_bh, zeta_prime_bh = rotate_points_with_quaternion(quaternion, [xi_prime_bh,y_prime_bh,zeta_prime_bh])
    xi_prime_bc, y_prime_bc, zeta_prime_bc = rotate_points_with_quaternion(quaternion, [xi_prime_bc,y_prime_bc,zeta_prime_bc])
    xi_prime_b2, y_prime_b2, zeta_prime_b2 = rotate_points_with_quaternion(quaternion, [xi_prime_b2,y_prime_b2,zeta_prime_b2])
                                                                                                                             
    xi_prime_ch, y_prime_ch, zeta_prime_ch = rotate_points_with_quaternion(quaternion, [xi_prime_ch,y_prime_ch,zeta_prime_ch])
    xi_prime   , y_prime   , zeta_prime    = rotate_points_with_quaternion(quaternion, [xi_prime   ,y_prime   ,zeta_prime   ])
                                                                                                                             
    # flip over y = z again after deflecting---------------------------------------------------------------
    
    y_prime_a1, zeta_prime_a1 = flip_2(y_prime_a1, zeta_prime_a1, vertical_wing, inverted_wing)
    y_prime_ah, zeta_prime_ah = flip_2(y_prime_ah, zeta_prime_ah, vertical_wing, inverted_wing)
    y_prime_ac, zeta_prime_ac = flip_2(y_prime_ac, zeta_prime_ac, vertical_wing, inverted_wing)
    y_prime_a2, zeta_prime_a2 = flip_2(y_prime_a2, zeta_prime_a2, vertical_wing, inverted_wing)
    y_prime_b1, zeta_prime_b1 = flip_2(y_prime_b1, zeta_prime_b1, vertical_wing, inverted_wing)
    y_prime_bh, zeta_prime_bh = flip_2(y_prime_bh, zeta_prime_bh, vertical_wing, inverted_wing)
    y_prime_bc, zeta_prime_bc = flip_2(y_prime_bc, zeta_prime_bc, vertical_wing, inverted_wing)
    y_prime_b2, zeta_prime_b2 = flip_2(y_prime_b2, zeta_prime_b2, vertical_wing, inverted_wing)
    y_prime_ch, zeta_prime_ch = flip_2(y_prime_ch, zeta_prime_ch, vertical_wing, inverted_wing)
    y_prime   , zeta_prime    = flip_2(y_prime,    zeta_prime   , vertical_wing, inverted_wing)

    # Pack the VD
    raw_VD.xi_prime_a1   = xi_prime_a1      
    raw_VD.xi_prime_ac   = xi_prime_ac      
    raw_VD.xi_prime_ah   = xi_prime_ah      
    raw_VD.xi_prime_a2   = xi_prime_a2      
    raw_VD.y_prime_a1    = y_prime_a1       
    raw_VD.y_prime_ah    = y_prime_ah       
    raw_VD.y_prime_ac    = y_prime_ac       
    raw_VD.y_prime_a2    = y_prime_a2       
    raw_VD.zeta_prime_a1 = zeta_prime_a1    
    raw_VD.zeta_prime_ah = zeta_prime_ah    
    raw_VD.zeta_prime_ac = zeta_prime_ac    
    raw_VD.zeta_prime_a2 = zeta_prime_a2    
                                            
    raw_VD.xi_prime_b1   = xi_prime_b1      
    raw_VD.xi_prime_bh   = xi_prime_bh      
    raw_VD.xi_prime_bc   = xi_prime_bc      
    raw_VD.xi_prime_b2   = xi_prime_b2      
    raw_VD.y_prime_b1    = y_prime_b1       
    raw_VD.y_prime_bh    = y_prime_bh       
    raw_VD.y_prime_bc    = y_prime_bc       
    raw_VD.y_prime_b2    = y_prime_b2       
    raw_VD.zeta_prime_b1 = zeta_prime_b1    
    raw_VD.zeta_prime_bh = zeta_prime_bh    
    raw_VD.zeta_prime_bc = zeta_prime_bc    
    raw_VD.zeta_prime_b2 = zeta_prime_b2    
                                            
    raw_VD.xi_prime_ch   = xi_prime_ch      
    raw_VD.xi_prime      = xi_prime         
    raw_VD.y_prime_ch    = y_prime_ch       
    raw_VD.y_prime       = y_prime          
    raw_VD.zeta_prime_ch = zeta_prime_ch    
    raw_VD.zeta_prime    = zeta_prime       

    
    return raw_VD    

# ----------------------------------------------------------------------
#  Make Hinge Quaternion
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
def make_hinge_quaternion(point_on_line, direction_unit_vector, rotation_angle):
    """ This make a quaternion that will rotate a vector about a the line that 
    passes through the point 'point_on_line' and has direction 'direction_unit_vector'.
    The quat rotates 'rotation_angle' radians. The quat is meant to be multiplied by
    the vector [x  y  z  1]

    Assumptions: 
    None

    Source:   
    https://sites.google.com/site/glennmurray/Home/rotation-matrices-and-formulas/rotation-about-an-arbitrary-axis-in-3-dimensions
    
    Inputs:   
    point_on_line         - a list or array of size 3 corresponding to point coords (a,b,c)
    direction_unit_vector - a list or array of size 3 corresponding to unit vector  <u,v,w>
    rotation_angle        - angle of rotation in radians
    n_points              - number of points that will be rotated
    
    Properties Used:
    N/A
    """       
    a,  b,  c  = point_on_line
    u,  v,  w  = direction_unit_vector
    
    cos         = np.cos(rotation_angle)
    sin         = np.sin(rotation_angle)
    
    q11 = u**2 + (v**2 + w**2)*cos
    q12 = u*v*(1-cos) - w*sin
    q13 = u*w*(1-cos) + v*sin
    q14 = (a*(v**2 + w**2) - u*(b*v + c*w))*(1-cos)  +  (b*w - c*v)*sin
    
    q21 = u*v*(1-cos) + w*sin
    q22 = v**2 + (u**2 + w**2)*cos
    q23 = v*w*(1-cos) - u*sin
    q24 = (b*(u**2 + w**2) - v*(a*u + c*w))*(1-cos)  +  (c*u - a*w)*sin
    
    q31 = u*w*(1-cos) - v*sin
    q32 = v*w*(1-cos) + u*sin
    q33 = w**2 + (u**2 + v**2)*cos
    q34 = (c*(u**2 + v**2) - w*(a*u + b*v))*(1-cos)  +  (a*v - b*u)*sin    
    
    quat = np.array([[q11, q12, q13, q14],
                     [q21, q22, q23, q24],
                     [q31, q32, q33, q34],
                     [0. , 0. , 0. , 1. ]])
    
    return quat

# ----------------------------------------------------------------------
#  Rotate Points with Quaternion
# ----------------------------------------------------------------------

## @ingroup Methods-Aerodynamics-Common-Fidelity_Zero-Lift
def rotate_points_with_quaternion(quat, points):
    """ This rotates the points by a quaternion

    Assumptions: 
    None

    Source:   
    https://sites.google.com/site/glennmurray/Home/rotation-matrices-and-formulas/rotation-about-an-arbitrary-axis-in-3-dimensions
    
    Inputs:   
    quat     - a quaternion that will rotate the given points about a line which 
               is not necessarily at the origin  
    points   - a list or array of size 3 corresponding to the lists (xs, ys, zs)
               where xs, ys, and zs are the (x,y,z) coords of the points 
               that will be rotated
    
    Outputs:
    xs, ys, zs - np arrays of the rotated points' xyz coordinates
    
    Properties Used:
    N/A
    """     
    vectors = np.array([points[0],points[1],points[2],np.ones(len(points[0]))]).T
    x_primes, y_primes, z_primes = np.sum(quat[0]*vectors, axis=1), np.sum(quat[1]*vectors, axis=1), np.sum(quat[2]*vectors, axis=1)
    return x_primes, y_primes, z_primes
    
    
    
def flip_1(A,B,T1,T2):
    """ This swaps values based on a double boolean
    
    If T1 is false A and B are kept as is. T2 is not used
    
    If T1 is true A and B are swapped
    T2 is applied at the first position

    Assumptions: 
    None

    Source:   
    N/A
    
    
    Inputs:   
    input_A  [int,float,array]
    input_B  [int,float,array]
    T1       [bool]
    T2       [bool/int/float]

    Outputs:
    swapped or not values
    
    Properties Used:
    N/A
    """
    
    NT1 = 1-T1
    
    return B*T1*T2+A*NT1, A*T1+B*NT1
    
def flip_2(A,B,T1,T2):
    """ This swaps values based on a double boolean
    
    If T1 is false A and B are kept as is. T2 is not used
    
    If T1 is true A and B are swapped
    T2 is applied at the second position

    Assumptions: 
    None

    Source:   
    N/A
    
    
    Inputs:   
    input_A  [int,float,array]
    input_B  [int,float,array]
    T1       [bool]
    T2       [bool/int/float]

    Outputs:
    swapped or not values
    
    Properties Used:
    N/A
    """
    NT1 = 1-T1
    
    return B*T1+A*NT1, A*T1*T2+B*NT1
    



# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/velocity_distribution.py
# 
# 
# Created:  Dec 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# pacakge imports  
import numpy as np  

# ---------------------------------------------------------------------------------------------------------------------- 
# velocity_distribution
# ---------------------------------------------------------------------------------------------------------------------- 
def velocity_distribution(qg,x,y,xbar,ybar,st,ct,alpha_2d,npanel,ncases,ncpts):
    """Compute the tangential velocity distribution at the       
                 midpoint of each panel   
    
    Source:
    None

    Assumptions:
    None  
    
    Inputs:                                                    

     qg          -  Vector of source/sink and vortex strengths    [unitless]          
     x           -  Vector of x coordinates of the surface nodes  [unitless]         
     y           -  Vector of y coordinates of the surface nodes  [unitless]            
     xbar        -  x-coordinate of the midpoint of each panel    [unitless]           
     ybar        -  y-coordinate of the midpoint of each panel    [unitless]           
     st          -  np.sin(theta) for each panel                  [radians]                
     ct          -  np.cos(theta) for each panel                  [radians]             
     al          -  Angle of attack in radians                    [radians]             
     npanel      -  Number of panels on the airfoil               [unitless]  

     Outputs:                                                        

     vt_2d        -  Vector of tangential velocities      

    Properties Used:
    N/A
    """   
    # flow tangency boundary condition - source distribution  
    vt_2d = ct *np.cos(alpha_2d) + st*np.sin(alpha_2d)
    gamma = np.repeat(qg[-1,:,:][np.newaxis,:,:],npanel, axis = 0)  
    
    # convert 1d matrices to 2d 
    qg_2d                = np.repeat(qg[:-1,:,:][np.newaxis,:,:,:],npanel, axis = 0) 
    x_2d                 = np.repeat(np.swapaxes(np.swapaxes(x,0, 2),0,1)[:,:,np.newaxis,:],npanel, axis = 2)
    y_2d                 = np.repeat(np.swapaxes(np.swapaxes(y,0, 2),0,1)[:,:,np.newaxis,:],npanel, axis = 2)
    xbar_2d              = np.repeat(np.swapaxes(np.swapaxes(xbar,0, 2),0,1)[:,:,:,np.newaxis],npanel, axis = 3)
    ybar_2d              = np.repeat(np.swapaxes(np.swapaxes(ybar,0, 2),0,1)[:,:,:,np.newaxis],npanel, axis = 3) 
    st_2d                = np.repeat(np.swapaxes(np.swapaxes(st,0, 2),0,1)[:,:,:,np.newaxis],npanel, axis = 3) 
    ct_2d                = np.repeat(np.swapaxes(np.swapaxes(ct,0, 2),0,1)[:,:,:,np.newaxis],npanel, axis = 3) 
    st_2d_T              = np.swapaxes(st_2d,2,3)
    ct_2d_T              = np.swapaxes(ct_2d,2,3)  
    
    # Fill the elements of the matrix of aero influence coefficients
    sti_minus_j          = ct_2d_T*st_2d - st_2d_T*ct_2d 
    cti_minus_j          = ct_2d_T*ct_2d + st_2d_T*st_2d 
    rij                  = np.sqrt((xbar_2d-x_2d[:,:,:,:-1])**2 + (ybar_2d-y_2d[:,:,:,:-1])**2)
    rij_plus_1           = np.sqrt((xbar_2d-x_2d[:,:,:,1:])**2 +  (ybar_2d-y_2d[:,:,:,1:])**2)
    rij_dot_rij_plus_1   = (xbar_2d-x_2d[:,:,:,:-1])*(xbar_2d-x_2d[:,:,:,1:]) + (ybar_2d-y_2d[:,:,:,:-1])*(ybar_2d-y_2d[:,:,:,1:])  
    anglesign            = np.sign((xbar_2d-x_2d[:,:,:,:-1])*(ybar_2d-y_2d[:,:,:,1:]) - (xbar_2d-x_2d[:,:,:,1:])*(ybar_2d-y_2d[:,:,:,:-1]))
    r_ratio              = rij_dot_rij_plus_1/rij/rij_plus_1
    r_ratio[r_ratio>1.0] = 1.0 # numerical noise     
    betaij               = np.real(anglesign*np.arccos(r_ratio))     
    # betaij[aoas,res,diag_indices,diag_indices] = np.pi
    for i in range(ncases):
        for j in range(ncpts):
            np.fill_diagonal(betaij[i,j,:,:], np.pi)   
    
    # swap axes 
    sti_minus_j_2d  = np.swapaxes(np.swapaxes(sti_minus_j,0,2),1,3) 
    betaij_2d       = np.swapaxes(np.swapaxes(betaij,0,2),1,3) 
    cti_minus_j_2d  = np.swapaxes(np.swapaxes(cti_minus_j,0,2),1,3) 
    rij_2d          = np.swapaxes(np.swapaxes(rij,0,2),1,3) 
    rij_plus_1_2d   = np.swapaxes(np.swapaxes(rij_plus_1,0,2),1,3) 
     
    vt_2d += np.sum(qg_2d/2/np.pi*(sti_minus_j_2d*betaij_2d - cti_minus_j_2d*np.log(rij_plus_1_2d/rij_2d)),1)  + \
             np.sum(gamma/2/np.pi*(sti_minus_j_2d*np.log(rij_plus_1_2d/rij_2d) + cti_minus_j_2d*betaij_2d),1)
    
    return  vt_2d



# RCAIDE/Methods/Energy/Propulsors/design_prop_rotor.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports    
from RCAIDE.Framework.Optimization.Packages.scipy                                              import scipy_setup 
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Design.optimization_setup       import optimization_setup
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Design.set_optimized_parameters import set_optimized_parameters

# Python package imports   
import time 

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Prop-rotor
# ----------------------------------------------------------------------------------------------------------------------   
def design_prop_rotor(rotor,number_of_stations = 20,solver_name= 'SLSQP',iterations = 200,
                      solver_sense_step = 1E-6,solver_tolerance = 1E-5,print_iterations = False):  
    """ Optimizes prop-rotor chord and twist given input parameters to meet either design power or thurst. 
        This scrip adopts RCAIDE's native optimization style where the objective function is expressed 
        as an aeroacoustic function, considering both efficiency and radiated noise.
          
          Inputs: 
          prop_attributes.
              hub radius                             [m]
              tip radius                             [m]
              rotation rate                          [rad/s]
              freestream velocity                    [m/s]
              number of blades                       [None]       
              number of stations                     [None]
              design lift coefficient                [None]
              airfoil data                           [None]
              optimization_parameters.      
                 slack_constaint                     [None]
                 ideal_SPL_dbA                       [dBA]
                 multiobjective_aeroacoustic_weight  [None]
            
          Outputs:
          Twist distribution                   [array of radians]
          Chord distribution                   [array of meters]
              
          Assumptions: 
             N/A 
        
          Source:
             None 
    """    

    # Unpack rotor geometry  
    rotor_tag     = rotor.tag
    rotor.tag     = 'rotor'
    
    # start optimization 
    ti                   = time.time()   
    optimization_problem = optimization_setup(rotor,number_of_stations,print_iterations)
    output               = scipy_setup.SciPy_Solve(optimization_problem,solver=solver_name, iter = iterations , sense_step = solver_sense_step,tolerance = solver_tolerance)    
    tf                   = time.time()
    elapsed_time         = round((tf-ti)/60,2)
    print('Lift-rotor Optimization Simulation Time: ' + str(elapsed_time) + ' mins')   
    
    # print optimization results 
    print (output)  
    
    # set remaining rotor variables using optimized parameters 
    rotor     = set_optimized_parameters(rotor,optimization_problem)
    rotor.tag = rotor_tag
     
    return rotor
   


# RCAIDE/Methods/Noise/Correlation_Buildup/Airframe/clean_wing_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

import numpy as np
from RCAIDE.Framework.Core import Units

# ----------------------------------------------------------------------------------------------------------------------
# Clean Wing Noise
# ----------------------------------------------------------------------------------------------------------------------
def clean_wing_noise(S,b,ND,IsHorz,velocity,viscosity,M,phi,theta,distance,frequency):
    """ This computes the 1/3 octave band sound pressure level and the overall sound pressure level from the clean wing,
    for a wing with area S (sq.ft) and span b (ft).  ND is a constant set to 0 for clean wings and set to 1 for propeller
    airplanes, jet transports with numerous large trailing edge flap tracks, flaps extended, or slats extended. ISHORZ must be set to 1.
    This function can be used for the horizontal tail by inserting the appropriate tail area and span. For a vertical tail, its appropriate
    area and height are used and ISHORZ must be set to 0.
    
    Assumptions:
        Correlation based.  
    
    Source:
       SAE Model
       
    Inputs:
            S                          - Wing Area [sq.ft]
            b                          - Wing Span [ft]
            ND                         - Costant from the method
            IsHoriz                    - Costant from the method
            deltaw                     - Wing Turbulent Boundary Layer thickness [ft]
            velocity                   - Aircraft speed [kts]
            viscosity                  - Dynamic viscosity
            M                          - Mach number
            phi                        - Azimuthal angle [rad]
            theta                      - Polar angle [rad]
            distance                   - Distance from airplane to observer, evaluated at retarded time [ft]
            frequency                  - Frequency array [Hz]



    Outputs: One Third Octave Band SPL [dB]
        SPL                              - Sound Pressure Level of the clean wing [dB]
        OASPL                            - Overall Sound Pressure Level of the clean wing [dB]

    Properties Used:
        None
    
    """
 
    delta  = 0.37*(S/b)*(velocity/Units.ft*S/(b*viscosity))**(-0.2)

    if IsHorz==1:
        DIR = np.cos(phi)
    elif IsHorz==0:
        DIR = np.sin(phi)


    if DIR==0:
        SPL = np.zeros(24)
    else:

        fmax  = 0.1*(velocity/Units.ft)/(delta*(1-M*np.cos(theta))) 

        OASPL = 50*np.log10((velocity/Units.kts)/100.0)+10*np.log10(delta*b/(distance**2.0))+8*ND+ \
            20*np.log10(DIR*np.sin(theta)*np.cos(theta/2.0))+104.3

        SPL   = OASPL+10.0*np.log10(0.613*(frequency/fmax)**4*((frequency/fmax)**1.5+0.5)**(-4))-0.03*np.abs(((frequency/fmax)-1))**1.5

    return SPL



# RCAIDE/Library/Plots/Energy/plot_battery_ragone_diagram.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_battery_ragone_diagram(battery,
                          save_figure   = False, 
                          save_filename = "Ragone_Plot",
                          file_type     =  ".png",
                          width = 11, height = 7):
    """Plots the pack-level conditions of the battery throughout flight.

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.
        freestream.altitude
        weights.total_mass
        weights.vehicle_mass_rate
        frames.body.thrust_force_vector

    Outputs:
    Plots

    Properties Used:
    N/A
    """
 
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
      
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height)  
        
    axis_1 = plt.subplot(1,1,1)
    esp_plot = np.linspace(battery.cell.ragone.lower_bound, battery.cell.ragone.upper_bound,50)
    psp_plot = battery.cell.ragone.const_1*10**(esp_plot*battery.cell.ragone.const_2)
    axis_1.plot(esp_plot/(Units.Wh/Units.kg),psp_plot/(Units.kW/Units.kg), color = 'black', marker = ps.markers[0], linewidth = ps.line_width, label= battery.tag) 
    axis_1.set_ylabel('Specific Power (kW/kg)')
    axis_1.set_xlabel('Specific Energy (W-h/kg)')
    set_axes(axis_1)       
     
    # set title of plot 
    title_text    = 'Battery Ragone Plot'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return  fig 


# RCAIDE/Library/Methods/Propulsors/Converters/supersonic_nozzle/compute_limited_geometry.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------     

# package imports
import numpy as np
from RCAIDE.Library.Methods.Gas_Dynamics.fm_id import fm_id
from RCAIDE.Library.Methods.Gas_Dynamics.nozzle_calculations import  pressure_ratio_isentropic, pressure_ratio_shock_in_nozzle


# ---------------------------------------------------------------------------------------------------------------------- 
# compute_limited_geometry
# ----------------------------------------------------------------------------------------------------------------------
def compute_limited_geometry(supersonic_nozzle, s_nozzle_conditions,conditions): 
    
    #unpack from conditions
    gamma    = conditions.freestream.isentropic_expansion_factor
    Cp       = conditions.freestream.specific_heat_at_constant_pressure
    Po       = conditions.freestream.pressure 
    R        = conditions.freestream.gas_specific_constant 
    
    #unpack from inputs
    Tt_in    = s_nozzle_conditions.inputs.stagnation_temperature
    Pt_in    = s_nozzle_conditions.inputs.stagnation_pressure                
    
    #unpack from self
    pid             = supersonic_nozzle.pressure_ratio
    etapold         = supersonic_nozzle.polytropic_efficiency
    max_area_ratio  = supersonic_nozzle.max_area_ratio
    min_area_ratio  = supersonic_nozzle.min_area_ratio
    
    
    # Method for computing the nozzle properties
    #--Getting the output stagnation quantities
    Pt_out   = Pt_in*pid
    Tt_out   = Tt_in*pid**((gamma-1.)/(gamma)*etapold)
    ht_out   = Cp*Tt_out

    # Method for computing the nozzle properties
    #-- Initial estimate for exit area
    area_ratio = (max_area_ratio + min_area_ratio)/2.
    
    #-- Compute limits of each possible flow condition       
    subsonic_pressure_ratio     = pressure_ratio_isentropic(area_ratio, gamma, True)
    nozzle_shock_pressure_ratio = pressure_ratio_shock_in_nozzle(area_ratio, gamma) 
    supersonic_max_Area         = pressure_ratio_isentropic(max_area_ratio, gamma, False)
    supersonic_min_Area         = pressure_ratio_isentropic(min_area_ratio, gamma, False)

    #-- Compute the output Mach number guess with freestream pressure
    #-- Initializing arrays
    P_out       = np.ones_like(Pt_out)
    A_ratio     = area_ratio*np.ones_like(Pt_out)
    M_out       = np.ones_like(Pt_out)

    # Establishing a correspondence between real pressure ratio and limits of each flow condition
    
    # Determine if flow is within subsonic/sonic range
    i_sub               = Po/Pt_out >= subsonic_pressure_ratio 
    
    # Detemine if there is a shock in nozzle
    i2                  = Po/Pt_out < subsonic_pressure_ratio
    i3                  = Po/Pt_out >= nozzle_shock_pressure_ratio
    i_shock             = np.logical_and(i2,i3)      
    
    # Determine if flow is overexpanded
    i4                  = Po/Pt_out < nozzle_shock_pressure_ratio
    i5                  = Po/Pt_out > supersonic_min_Area
    i_over              = np.logical_and(i4,i5)  
    
    # Determine if flow is supersonic
    i6                  = Po/Pt_out <= supersonic_min_Area
    i7                  = Po/Pt_out >= supersonic_max_Area
    i_sup               = np.logical_and(i6,i7) 
    
    # Determine if flow is underexpanded
    i_und               = Po/Pt_out < supersonic_max_Area
    
    #-- Subsonic and sonic flow
    P_out[i_sub]        = Po[i_sub]
    M_out[i_sub]        = np.sqrt((((Pt_out[i_sub]/P_out[i_sub])**((gamma[i_sub]-1.)/gamma[i_sub]))-1.)*2./(gamma[i_sub]-1.))
    A_ratio[i_sub]      = 1./fm_id(M_out[i_sub],gamma[i_sub])
    
    #-- Shock inside nozzle
    P_out[i_shock]      = Po[i_shock]
    M_out[i_shock]      = np.sqrt((((Pt_out[i_shock]/P_out[i_shock])**((gamma[i_shock]-1.)/gamma[i_shock]))-1.)*2./(gamma[i_shock]-1.))   
    A_ratio[i_shock]    = 1./fm_id(M_out[i_shock],gamma[i_shock])
    
    #-- Overexpanded flow
    P_out[i_over]       = supersonic_min_Area[i_over]*Pt_out[i_over] 
    M_out[i_over]       = np.sqrt((((Pt_out[i_over]/P_out[i_over])**((gamma[i_over]-1.)/gamma[i_over]))-1.)*2./(gamma[i_over]-1.))
    A_ratio[i_over]     = 1./fm_id(M_out[i_over],gamma[i_over])
    
    #-- Isentropic supersonic flow, with variable area adjustments
    P_out[i_sup]        = Po[i_sup]
    M_out[i_sup]        = np.sqrt((((Pt_out[i_sup]/P_out[i_sup])**((gamma[i_sup]-1.)/gamma[i_sup]))-1.)*2./(gamma[i_sup]-1.))    
    A_ratio[i_sup]      = 1./fm_id(M_out[i_sup],gamma[i_sup])
    
    #-- Underexpanded flow
    P_out[i_und]        = supersonic_max_Area[i_und]*Pt_out[i_und] 
    M_out[i_und]        = np.sqrt((((Pt_out[i_und]/P_out[i_und])**((gamma[i_und]-1.)/gamma[i_und]))-1.)*2./(gamma[i_und]-1.))
    A_ratio[i_und]      = 1./fm_id(M_out[i_und],gamma[i_und])
    
    #-- Calculate other flow properties
    T_out   = Tt_out/(1.+(gamma-1.)/2.*M_out*M_out)
    h_out   = Cp*T_out
    u_out   = M_out*np.sqrt(gamma*R*T_out)
    rho_out = P_out/(R*T_out)

    #pack computed quantities into outputs
    s_nozzle_conditions.outputs.stagnation_temperature  = Tt_out
    s_nozzle_conditions.outputs.stagnation_pressure     = Pt_out
    s_nozzle_conditions.outputs.stagnation_enthalpy     = ht_out
    s_nozzle_conditions.outputs.mach_number             = M_out
    s_nozzle_conditions.outputs.static_temperature      = T_out
    s_nozzle_conditions.outputs.rho                     = rho_out
    s_nozzle_conditions.outputs.static_enthalpy         = h_out
    s_nozzle_conditions.outputs.velocity                = u_out
    s_nozzle_conditions.outputs.static_pressure         = P_out
    s_nozzle_conditions.outputs.area_ratio              = A_ratio

    return 
    




# RCAIDE/Framework/Optmizaition/Common/Nexus.py 

# ----------------------------------------------------------------------------------------------------------------- 
#  IMPORT
# --- ------------------------------------------------------------------------------------------------------------- 
 
import RCAIDE
from RCAIDE.Framework.Core import Data
from RCAIDE.Framework.Analyses import Process
from copy import deepcopy

from . import helper_functions as help_fun
import numpy as np

# ----------------------------------------------------------------------------------------------------------------- 
#  Nexus Class
# --- ------------------------------------------------------------------------------------------------------------- 
## @ingroupFramework-Optimization-Common
class Nexus(Data):
    """noun (plural same or nexuses)
        -a connection or series of connections linking two or more things
        -a connected group or series: a nexus of ideas.
        -the central and most important point or place
        
        This is the class that makes optimization possible. We put all the data and functions together to make
        your future dreams come true.
         
    """    
    
    def __defaults__(self):
        """This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        self.vehicle_configurations = RCAIDE.Library.Components.Configs.Config.Container()
        self.analyses               = RCAIDE.Framework.Analyses.Analysis.Container()
        self.missions               = None
        self.procedure              = Process()
        self.results                = Data()
        self.summary                = Data()
        self.optimization_problem   = None
        self.fidelity_level         = 1
        self.last_inputs            = None
        self.last_fidelity          = None
        self.evaluation_count       = 0
        self.force_evaluate         = False
        self.hard_bounded_inputs    = False
    
    def evaluate(self,x = None):
        """This function runs the problem you setup in RCAIDE
            If the last time you ran this the inputs were the same, a cache is used.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            x       [vector]
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        
        self.unpack_inputs(x)
        
        # Check if last call was the same
        if np.all(self.optimization_problem.inputs==self.last_inputs) \
           and self.last_fidelity == self.fidelity_level \
           and self.force_evaluate == False:
            pass
        else:
            self._really_evaluate()
        
    
    def _really_evaluate(self):
        """Tricky little function you're not supposed to use. Doesn't check if the last inputs were already run.
            This steps through like a process through the nexus, and stores the results.
    
            Assumptions:
            Doesn't set values!
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        
        nexus = self
        
        self.evaluation_count += 1
        
        for key,step in nexus.procedure.items():
            if hasattr(step,'evaluate'):
                self = step.evaluate(nexus)
            else:
                nexus = step(nexus)
            self = nexus
                
        # Store to cache
        self.last_inputs   = deepcopy(self.optimization_problem.inputs)
        self.last_fidelity = self.fidelity_level
          
    
    def objective(self,x = None):
        """Retrieve the objective value for your function
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x       [vector]
    
            Outputs:
            scaled_objective [float]
    
            Properties Used:
            None
        """           
    
        self.evaluate(x)
        
        aliases     = self.optimization_problem.aliases
        objective   = self.optimization_problem.objective
    
        objective_value  = help_fun.get_values(self,objective,aliases)  
        scaled_objective = help_fun.scale_obj_values(objective,objective_value)
        
        return scaled_objective.astype(np.double) 
    
    def inequality_constraint(self,x = None):
        """Retrieve the inequality constraint values for your function
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x                  [vector]
    
            Outputs:
            scaled_constraints [vector]
    
            Properties Used:
            None
            """           
        
        self.evaluate(x)
        
        aliases     = self.optimization_problem.aliases
        constraints = self.optimization_problem.constraints 
        
        # Setup constraints  
        indices = []
        for ii in range(0,len(constraints)):
            if constraints[ii][1]==('='):
                indices.append(ii)        
        iqconstraints = np.delete(constraints,indices,axis=0)
    
        if len(iqconstraints) == 0:
            scaled_constraints = []
        else:

            # get constaint values 
            constraint_values = help_fun.get_values(self,iqconstraints,aliases)          
            
            # scale bounds 
            scaled_bnd_constraints  = help_fun.scale_const_bnds(iqconstraints)
            
            # scale constaits 
            scaled_constraints = help_fun.scale_const_values(iqconstraints,constraint_values)
            
            # determine difference between bounds and constaints 
            constraint_evaluations = scaled_constraints  - scaled_bnd_constraints
            
            # coorect constaints based on sign 
            constraint_evaluations[iqconstraints[:,1]=='<'] = -constraint_evaluations[iqconstraints[:,1]=='<']
            
        return constraint_evaluations       
    
    def equality_constraint(self,x = None):
        """Retrieve the equality constraint values for your function
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x                  [vector]
    
            Outputs:
            scaled_constraints [vector]
    
            Properties Used:
            None
        """         
    
        self.evaluate(x)

        aliases     = self.optimization_problem.aliases
        constraints = self.optimization_problem.constraints
        
        # Setup constraints  
        indices = []
        for ii in range(0,len(constraints)):
            if constraints[ii][1]=='>':
                indices.append(ii)
            elif constraints[ii][1]=='<':
                indices.append(ii)
        eqconstraints = np.delete(constraints,indices,axis=0)
    
        if len(eqconstraints) == 0:
            scaled_constraints = []
        else:
            constraint_values  = help_fun.get_values(self,eqconstraints,aliases)
            scaled_constraints = help_fun.scale_const_values(eqconstraints,constraint_values) - help_fun.scale_const_bnds(eqconstraints)

        return scaled_constraints   
        
    def all_constraints(self,x = None):
        """Returns both the inequality and equality constraint values for your function
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x                  [vector]
    
            Outputs:
            scaled_constraints [vector]
    
            Properties Used:
            None
        """         
        
        self.evaluate(x)
        
        aliases     = self.optimization_problem.aliases
        constraints = self.optimization_problem.constraints
    
        constraint_values  = help_fun.get_values(self,constraints,aliases) 
        scaled_constraints = help_fun.scale_const_values(constraints,constraint_values) 

        return scaled_constraints     
    
    
    def unpack_inputs(self,x = None):
        """Put's the values of the problem in the right place.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x                  [vector]
    
            Outputs:
            None
    
            Properties Used:
            None
        """                 
         
        # Scale the inputs if given
        inputs = self.optimization_problem.inputs
        if x is not None:
            inputs = help_fun.scale_input_values(inputs,x)
            
        # Limit the values to the edges
        if self.hard_bounded_inputs:
            inputs = help_fun.limit_input_values(inputs)        
            
        # Convert units
        converted_values = help_fun.convert_values(inputs)

        # Set the dictionary
        aliases = self.optimization_problem.aliases
        
        self    = help_fun.set_values(self,inputs,converted_values,aliases)     

    
    def constraints_individual(self,x = None):
        """Put's the values of the problem in the right place.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x                  [vector]
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        pass     

    def finite_difference(self,x,diff_interval=1e-8):
        """Finite difference gradients and jacobians of the problem.
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x                  [vector]
            diff_interval      [float]
    
            Outputs:
            grad_obj           [vector]
            jac_con            [array]
    
            Properties Used:
            None
        """           
        
        obj = self.objective(x)
        con = self.all_constraints(x)
        
        inpu  = self.optimization_problem.inputs
        const = self.optimization_problem.constraints
        
        inplen = len(inpu)
        conlen = len(const)
        
        grad_obj = np.zeros(inplen)
        jac_con  = np.zeros((inplen,conlen))
        
        con2 = (con*np.ones_like(jac_con))
        
        for ii in range(0,inplen):
            newx     = np.asarray(x)*1.0
            newx[ii] = newx[ii] + diff_interval
            
            grad_obj[ii]  = self.objective(newx)
            jac_con[ii,:] = self.all_constraints(newx)
        
        grad_obj = (grad_obj - obj)/diff_interval
        
        jac_con = (jac_con - con2).T/diff_interval
        
        grad_obj = grad_obj.astype(float)
        jac_con  = jac_con.astype(float)
        
        return grad_obj, jac_con
    
    
    def translate(self,x = None):
        """Make a pretty table view of the problem with objective and constraints at the current inputs
    
            Assumptions:
            N/A
    
            Source:
            N/A
    
            Inputs:
            x                  [vector]
    
            Outputs:
            inpu               [array]
            const_table        [array]
    
            Properties Used:
            None
        """         
        
        # Run the problem just in case
        self.evaluate(x)
        
        # Pull out the inputs and print them
        inpu       = self.optimization_problem.inputs
        print('Design Variable Table:\n')
        print(inpu)
        
        # Print the objective value
        obj         = self.optimization_problem.objective
        obj_val     = self.objective(x)
        obj_scale   = help_fun.unscale_const_values(obj,obj_val)
        obj_table   = np.array(obj)
        obj_table   = np.insert(obj_table,1,obj_scale)
        
        print('\nObjective Table:\n')
        print(obj_table)
        
        # Pull out the constraints
        const       = self.optimization_problem.constraints
        const_vals  = self.all_constraints(x)
        const_scale = help_fun.unscale_const_values(const,const_vals)
        
        # Make a new table
        const_table = np.array(const)
        const_table = np.insert(const_table,1,const_scale,axis=1)

        print('\nConstraint Table:\n')
        print(const_table)
        
        return inpu,const_table
                               
        
    
 



# RCAIDE/Library/Missions/Common/Update/acceleration.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Update Acceleration
# ----------------------------------------------------------------------------------------------------------------------   
def acceleration(segment):
    """ Differentiates the velocity vector to get accelerations
    
        Assumptions:
        Assumes a flat earth, this is planar motion.
        
        Inputs:
            segment.state.conditions:
                frames.inertial.velocity_vector     [meters/second]
            segment.state.numerics.time.differentiate       [float]
            
        Outputs:
            segment.state.conditions:           
                frames.inertial.acceleration_vector [meters]

        Properties Used:
        N/A
                                
    """            
    
    # unpack conditions
    v = segment.state.conditions.frames.inertial.velocity_vector
    D = segment.state.numerics.time.differentiate
    
    # accelerations
    acc = np.dot(D,v)
    
    # pack conditions
    segment.state.conditions.frames.inertial.acceleration_vector[:,:] = acc[:,:]   